from enum import Enum, auto

from .routes import *


class AccessType(Enum):
    """Enumeration of access types. Specifies how data will be read
    and returned.

    * JSON_STR: Result will be returned as an unparsed json string.

    * FILE_PATH: Result will be returned as the file path to the file
    containing the data.

    * OBJ: The json will be parsed and returned as a python object.

    * URI: A string which is a unique identifier of this asset between
    implementations of Aerovaldb. Can be used with :meth:`aerovaldb.AerovalDB.get_by_uuid`
    and :meth:`aerovaldb.AerovalDB.put_by_uuid` to read or write respectively.

    * MTIME: The timestamp for last modification for the resource will be
    returned (as datetime.datetime).

    * CTIME: The creation timestamp for the resource will be returned (as
    datetime.datetime)
    """

    JSON_STR = auto()
    FILE_PATH = auto()
    OBJ = auto()
    URI = auto()
    BLOB = auto()
    MTIME = auto()
    CTIME = auto()



import asyncio
import fcntl
import logging
import pathlib
import sys
from abc import ABC, abstractmethod

from ..utils import has_async_loop, run_until_finished

if sys.version_info >= (3, 12):
    from typing import override
else:
    from typing_extensions import override

logger = logging.getLogger(__name__)


class AerovaldbLock(ABC):
    """
    Interface for a context manager based locking mechanism.

    Can be used as a context manager (ie. in a with block).
    """

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_value, exc_traceback):
        self.release()

    @abstractmethod
    def acquire(self):
        """
        Acquire the lock manually. Usually this should be done
        using a with statement.
        """
        pass

    @abstractmethod
    def release(self):
        """
        Release the lock manually. Usually this should be done
        using a with statement.
        """
        pass

    @abstractmethod
    def is_locked(self) -> bool:
        """
        Check whether the lock is currently acquired.
        """
        pass


class FakeLock(AerovaldbLock):
    """
    A lock class which does not lock anything. Useful for
    disabling locking when it is not really needed, but leave
    code written for using locking.
    """

    def __init__(self):
        logger.debug("Initializing FAKE lock")
        self.acquire()

    @override
    def acquire(self):
        self._acquired = True

    @override
    def release(self):
        self._acquired = False

    @override
    def is_locked(self) -> bool:
        return self._acquired


class FileLock(AerovaldbLock):
    def __init__(self, lock_file: str | pathlib.Path):
        logger.debug("Initializing lock with lockfile %s", lock_file)
        self._lock_file = lock_file
        self._lock_handle = open(lock_file, "a+")
        self._aiolock = asyncio.Lock()
        self.acquire()

    @override
    def acquire(self):
        logger.debug("Acquiring lock with lockfile %s", self._lock_file)

        if has_async_loop():
            run_until_finished(self._aiolock.acquire)

        fcntl.lockf(self._lock_handle, fcntl.LOCK_EX)
        self._acquired = True

    @override
    def release(self):
        logger.debug("Releasing lock with lockfile %s", self._lock_file)

        fcntl.fcntl(self._lock_handle, fcntl.LOCK_UN)
        self._acquired = False
        if self._aiolock.locked():
            self._aiolock.release()

    @override
    def is_locked(self) -> bool:
        return self._acquired



from enum import Enum

_ROUTE_GLOB_STATS = "/v0/glob_stats/{project}/{experiment}/{frequency}"

_ROUTE_REG_STATS = "/v0/regional_stats/{project}/{experiment}/{frequency}"

_ROUTE_HEATMAP = "/v0/heatmap/{project}/{experiment}/{frequency}"

_ROUTE_CONTOUR = "/v0/contour/{project}/{experiment}/{obsvar}/{model}"
_ROUTE_CONTOUR_TIMESPLIT = (
    "/v0/contour2/{project}/{experiment}/{obsvar}/{model}/{timestep}"
)

_ROUTE_TIMESERIES = (
    "/v0/ts/{project}/{experiment}/{location}/{network}/{obsvar}/{layer}"
)

_ROUTE_TIMESERIES_WEEKLY = (
    "/v0/ts_weekly/{project}/{experiment}/{location}/{network}/{obsvar}/{layer}"
)

_ROUTE_EXPERIMENTS = "/v0/experiments/{project}"

_ROUTE_CONFIG = "/v0/config/{project}/{experiment}"

_ROUTE_MENU = "/v0/menu/{project}/{experiment}"

_ROUTE_STATISTICS = "/v0/statistics/{project}/{experiment}"

_ROUTE_RANGES = "/v0/ranges/{project}/{experiment}"

_ROUTE_REGIONS = "/v0/regions/{project}/{experiment}"

_ROUTE_MODELS_STYLE = "/v0/model_style/{project}"

_ROUTE_MAP = (
    "/v0/map/{project}/{experiment}/{network}/{obsvar}/{layer}/{model}/{modvar}"
)

_ROUTE_SCATTER = (
    "/v0/scat/{project}/{experiment}/{network}/{obsvar}/{layer}/{model}/{modvar}"
)

_ROUTE_PROFILES = "/v0/profiles/{project}/{experiment}/{location}/{network}/{obsvar}"

_ROUTE_HEATMAP_TIMESERIES = "/v0/hm_ts/{project}/{experiment}"

_ROUTE_FORECAST = (
    "/v0/forecast/{project}/{experiment}/{region}/{network}/{obsvar}/{layer}"
)

_ROUTE_FAIRMODE = (
    "/v0/fairmode/{project}/{experiment}/{region}/{network}/{obsvar}/{layer}"
)

_ROUTE_GRIDDED_MAP = "/v0/gridded_map/{project}/{experiment}/{obsvar}/{model}"

_ROUTE_REPORT = "/v0/report/{project}/{experiment}/{title}"

_ROUTE_REPORT_IMAGE = "/v0/report-image/{project}/{experiment}/{path}"

_ROUTE_MAP_OVERLAY = "/v0/map-overlay/{project}/{experiment}/{source}/{variable}/{date}"


class Route(Enum):
    GLOB_STATS = _ROUTE_GLOB_STATS
    REGIONAL_STATS = _ROUTE_REG_STATS
    HEATMAP = _ROUTE_HEATMAP
    CONTOUR = _ROUTE_CONTOUR
    CONTOUR_TIMESPLIT = _ROUTE_CONTOUR_TIMESPLIT
    TIMESERIES = _ROUTE_TIMESERIES
    TIMESERIES_WEEKLY = _ROUTE_TIMESERIES_WEEKLY
    EXPERIMENTS = _ROUTE_EXPERIMENTS
    CONFIG = _ROUTE_CONFIG
    MENU = _ROUTE_MENU
    STATISTICS = _ROUTE_STATISTICS
    RANGES = _ROUTE_RANGES
    REGIONS = _ROUTE_REGIONS
    MODELS_STYLE = _ROUTE_MODELS_STYLE
    MAP = _ROUTE_MAP
    SCATTER = _ROUTE_SCATTER
    PROFILES = _ROUTE_PROFILES
    HEATMAP_TIMESERIES = _ROUTE_HEATMAP_TIMESERIES
    FORECAST = _ROUTE_FORECAST
    FAIRMODE = _ROUTE_FAIRMODE
    GRIDDED_MAP = _ROUTE_GRIDDED_MAP
    REPORT = _ROUTE_REPORT
    REPORT_IMAGE = _ROUTE_REPORT_IMAGE
    MAP_OVERLAY = _ROUTE_MAP_OVERLAY



from .asyncio import async_and_sync, has_async_loop, run_until_finished
from .json import json_dumps_wrapper
from .string_utils import str_to_bool
from .uri import (
    build_uri,
    decode_arg,
    encode_arg,
    extract_substitutions,
    parse_formatted_string,
    parse_uri,
)



import os

import pytest

import aerovaldb
from aerovaldb.jsondb.jsonfiledb import AerovalJsonFileDB
from aerovaldb.sqlitedb.sqlitedb import AerovalSqliteDB


def test_plugins():
    engines = aerovaldb.list_engines()
    print(engines)
    assert len(engines) == 2


def test_open_json_1():
    with aerovaldb.open("json_files:.") as db:
        assert isinstance(db, AerovalJsonFileDB)
        assert os.path.abspath(db._basedir) == os.path.abspath(".")


def test_open_json_2():
    with aerovaldb.open(".") as db:
        assert isinstance(db, AerovalJsonFileDB)
        assert os.path.abspath(db._basedir) == os.path.abspath(".")


@pytest.mark.parametrize(
    "fext", (pytest.param(".sqlite", id="sqlite"), pytest.param(".db", id="db"))
)
def test_open_sqlite_1(tmp_path, fext):
    path = os.path.join(tmp_path, f"test{fext}")
    with aerovaldb.open(f"sqlitedb:{path}") as db:
        assert isinstance(db, AerovalSqliteDB)
        assert db._dbfile == path


@pytest.mark.parametrize(
    "fext", (pytest.param(".sqlite", id="sqlite"), pytest.param(".db", id="db"))
)
def test_open_sqlite_2(tmp_path, fext):
    path = os.path.join(tmp_path, f"test{fext}")
    with aerovaldb.open(path) as db:
        assert isinstance(db, AerovalSqliteDB)
        assert db._dbfile == path


def test_open_sqlite_3():
    with aerovaldb.open(":memory:") as db:
        assert isinstance(db, AerovalSqliteDB)
        assert db._dbfile == ":memory:"



import re

PATH_COMPONENT_PATTERN = re.compile(r"^[^/]+$", flags=re.UNICODE)


def str_to_bool(value: str, /, default: bool | None = None) -> bool:
    """
    Parses a string as a boolean, supporting various values. It is intended
    mainly for parsing environment variables.

    Supports 1/0, true/false, t/f, yes/no, y/n (Case insensitive).

    :param value : The string value to be converted.
    :param default : Optional default return value (True/False) if the string
        value doesn't match any supported value. If not set, or set to None,
        a ValueError is raised in these cases.

    :raises ValueError
        If value is not a string
    :raises ValueError
        Raised on unsupported input value. Suppressed if default is set to True/False.

    >>> from aerovaldb.utils.string_utils import str_to_bool
    >>> str_to_bool("y")
    True
    >>> str_to_bool("blah", default=True)
    True
    >>> str_to_bool("n")
    False
    """
    if not isinstance(value, str):
        raise ValueError(f"Expected str, got {type(value)}")

    if value.lower() in ["1", "true", "t", "yes", "y"]:
        return True

    if value.lower() in ["0", "false", "f", "no", "n"]:
        return False

    if default is not None:
        return default

    raise ValueError("Could not convert string to bool: '{value}'")



import simplejson  # type: ignore


def json_encoder(obj):
    if isinstance(obj, set):
        return list(obj)

    TypeError(repr(obj) + " is not JSON serializable")


def json_dumps_wrapper(obj, **kwargs) -> str:
    """
    Wrapper which calls simplejson.dumps with the correct options, known to work for objects
    returned by Pyaerocom.

    This ensures that nan values are serialized as null to be compliant with the json standard.
    """
    return simplejson.dumps(obj, ignore_nan=True, default=json_encoder, **kwargs)



import pytest

from aerovaldb import Route
from aerovaldb.utils import (
    decode_arg,
    encode_arg,
    extract_substitutions,
    parse_formatted_string,
    parse_uri,
)


@pytest.mark.parametrize(
    "template,result",
    (
        ("{A}{B}{C}", {"A", "B", "C"}),
        ("{A}hello world{B} test {C}", {"A", "B", "C"}),
        ("", set()),
    ),
)
def test_extract_substitutions(template: str, result: set[str]):
    l = extract_substitutions(template)

    assert set(l) == result


@pytest.mark.parametrize(
    "template,string,expected",
    (
        ("{test}", "hello", {"test": "hello"}),
        ("ABCD{test}1234", "ABCDhelloworld1234", {"test": "helloworld"}),
        (
            "test/{a}/{b}/{c}/{d}",
            "test/A/B/C/D",
            {"a": "A", "b": "B", "c": "C", "d": "D"},
        ),
    ),
)
def test_parse_formatted_string(template: str, string: str, expected: dict):
    assert parse_formatted_string(template, string) == expected


@pytest.mark.parametrize(
    "template,string,val,exception",
    (
        ("{a}{b}", "abcd", "can not be disambiguated", Exception),
        (
            "{a}b{b}c",
            "testbhellotestblah",
            "did not match template string",
            Exception,
        ),
    ),
)
def test_parse_fromatted_string_error(template: str, string: str, val: str, exception):
    with pytest.raises(exception) as e:
        parse_formatted_string(template, string)

    assert val in str(e.value)


@pytest.mark.parametrize(
    "uri,expected",
    (
        (
            "/v0/experiments/project",
            (Route.EXPERIMENTS, {"project": "project"}, {}),
        ),
        (
            "/v0/map/project/experiment/network/obsvar/layer/model/modvar?time=time",
            (
                Route.MAP,
                {
                    "project": "project",
                    "experiment": "experiment",
                    "network": "network",
                    "obsvar": "obsvar",
                    "layer": "layer",
                    "model": "model",
                    "modvar": "modvar",
                },
                {"time": "time"},
            ),
        ),
    ),
)
def test_parse_uri(uri: str, expected: tuple[str, dict, dict]):
    template, route_args, kwargs = parse_uri(uri)

    assert (template, route_args, kwargs) == expected


def test_parse_uri_error():
    with pytest.raises(ValueError):
        parse_uri("??")


@pytest.mark.parametrize(
    "input,expected",
    (
        ("", ""),
        ("%", "%0"),
        ("/", "%1"),
        ("hello-world/hello%1234", "hello-world%1hello%01234"),
        ("%/" * 5, "%0%1" * 5),
    ),
)
def test_encode_decode_arg(input: str, expected: str):
    encoded = encode_arg(input)

    assert encoded == expected

    decoded = decode_arg(encoded)

    assert decoded == input



class UnusedArguments(ValueError):
    """
    Raised by jsondb if args or kwargs remain after matching (which
    is likely a mistake by the caller).
    """


class TemplateNotFound(KeyError):
    """
    Raised by jsondb if no matching filepath template was found for
    the requested version of data.
    """


class UnsupportedOperation(NotImplementedError):
    """
    Raised if some operation could not be provided for some reason.
    """



# These filters are used to filter end points return a subset of a file / endpoint.
# Currently this only applies to regional_stats and heatmap, which both return a
# subset of the globstats endpoint.


def filter_regional_stats(data, variable: str, network: str, layer: str):
    """
    Filters regional stats out of a glob_stats data object.

    :data : Data object to operate on.
    :variable : Variable name.
    :network : Observation network.
    :layer : Layer.
    """
    return data[variable][network][layer]


def filter_heatmap(data, region: str, time: str):
    """
    Filters heatmap data out of a glob stats data object.

    :data : Data object to operate on.
    :region : Region ID.
    :time : Time.
    """
    filtered_data = {}  # type: ignore
    for variable, variable_data in data.items():
        filtered_data.setdefault(variable, {})
        for network, network_data in variable_data.items():
            filtered_data[variable].setdefault(network, {})
            for layer, layer_data in network_data.items():
                filtered_data[variable][network].setdefault(layer, {})
                for model, model_data in layer_data.items():
                    filtered_data[variable][network][layer].setdefault(model, {})
                    for model_var, model_var_data in model_data.items():
                        filtered_data[variable][network][layer][model].setdefault(
                            model_var, {}
                        )
                        if region in model_var_data:
                            region_data = model_var_data[region]
                            if time in region_data:
                                filtered_data[variable][network][layer][model][
                                    model_var
                                ][region] = {time: region_data[time]}

    return filtered_data


def filter_contour(data, timestep: str | None = None):
    if timestep == None:
        return data

    return data[timestep]


def filter_map(data, frequency: str | None = None, season: str | None = None):
    if all([x is None for x in [frequency, season]]):
        return data

    if all([isinstance(x, str) for x in [frequency, season]]):
        keys_to_keep = {
            "station_name",
            "latitude",
            "longitude",
            "altitude",
            "region",
            frequency,
        }
        for item in data:
            for k1 in list(item.keys()):
                if k1 not in keys_to_keep:
                    del item[k1]
                    continue

                for k2 in list(item[frequency].keys()):
                    if k2 != season:
                        del item[frequency][k2]

        return data

    raise ValueError(
        f"frequency and season must either both be None, or both be provided. Got {[frequency, season]}"
    )



class EncodedStr(str):
    pass


class DecodedStr(str):
    pass


def encode_str(string: str, *, encode_chars: dict[str, str]) -> EncodedStr:
    """Encodes a string by replacing characters by en encoded
    character sequence.

    :param string: String to be encoded.
    :param encode_chars: Mapping from character to replacement string
    sequence in encoded string.

    :return: Encoded string.
    """
    if isinstance(string, EncodedStr):
        raise TypeError(f"String '{string}' is already encoded.")

    mapping = str.maketrans(encode_chars)
    return EncodedStr(string.translate(mapping))


def decode_str(string: str, *, encode_chars: dict[str, str]) -> DecodedStr:
    """Decodes a string previously encoded using encode_str.

    :param string: String to be decoded.
    :param encode_chars: Mapping from character to replacement string
    used during the original encode operation.
    :return: Decoded string.
    """
    if isinstance(string, DecodedStr):
        raise TypeError(f"String '{string}' is already decoded.")

    for k, v in encode_chars.items():
        string = string.replace(v, k)
    return DecodedStr(string)



# This test file applies tests to each implementation of database connector
# through the interface defined in src/aerovaldb/aerovaldb.py
# In addition each implementation has implementation specific tests in its
# respective test_*.py file
# - json_files: tests/jsondb/test_jsonfiledb.py
# - sqlitedb:   tests/sqlitedb/test_sqlitedb.py

import datetime
import pathlib
import random

import filetype
import pytest
import simplejson  # type: ignore

import aerovaldb
import aerovaldb.jsondb
from aerovaldb.utils.copy import copy_db_contents


@pytest.fixture
def tmpdb(tmp_path, dbtype: str) -> aerovaldb.AerovalDB:
    """Fixture encapsulating logic for each tested database connection to create
    a fresh, temporary database and connect to it."""
    if dbtype == "json_files":
        return aerovaldb.open(f"json_files:{str(tmp_path)}")
    elif dbtype == "sqlitedb":
        return aerovaldb.open(":memory:")

    assert False


TESTDB_PARAMETRIZATION = pytest.mark.parametrize(
    # This is a parametrization which returns the correct resource string to access
    # the prebuilt test database for each database connector.
    "testdb",
    (
        pytest.param(
            "json_files:./tests/test-db/json",
        ),
        pytest.param(
            "sqlitedb:./tests/test-db/sqlite/test.sqlite",
        ),
    ),
)

IMPLEMENTATION_PARAMETRIZATION = pytest.mark.parametrize(
    "dbtype", (pytest.param("json_files"), pytest.param("sqlitedb"))
)
GET_PARAMETRIZATION = pytest.mark.parametrize(
    "fun,args,kwargs,expected",
    (
        (
            "get_glob_stats",
            ["project", "experiment", "frequency"],
            None,
            "./project/experiment/hm/",
        ),
        (
            "get_regional_stats",
            ["project", "experiment", "frequency", "network", "variable", "layer"],
            None,
            "./project/experiment/hm/regional_stats",
        ),
        pytest.param(
            "get_heatmap",
            ["project", "experiment", "frequency", "region", "time"],
            None,
            "./project/experiment/hm/regional_stats",
            marks=pytest.mark.xfail(reason="missing test file in json testdb"),
        ),
        (
            "get_contour",
            ["project", "experiment", "modvar", "model"],
            {"timestep": "timestep"},
            "748956457892",
        ),
        (
            "get_contour",
            ["project", "experiment", "modvar", "model"],
            {"timestep": "timestep2"},
            "2758924570298570",
        ),
        (
            "get_timeseries",
            ["project", "experiment", "location", "network", "obsvar", "layer"],
            None,
            "./project/experiment/ts/",
        ),
        (
            "get_timeseries_weekly",
            ["project", "experiment", "location", "network", "obsvar", "layer"],
            None,
            "./project/experiment/ts/dirunal/",
        ),
        ("get_config", ["project", "experiment"], None, "./project/experiment/"),
        ("get_menu", ["project", "experiment"], None, "./project/experiment/"),
        ("get_statistics", ["project", "experiment"], None, "./project/experiment/"),
        ("get_ranges", ["project", "experiment"], None, "./project/experiment/"),
        ("get_regions", ["project", "experiment"], None, "./project/experiment/"),
        ("get_models_style", ["project"], None, "./project/"),
        ("get_experiments", ["project"], None, "./project/"),
        (
            "get_models_style",
            ["project"],
            {"experiment": "experiment"},
            "./project/experiment/",
        ),
        (
            "get_map",
            [
                "project",
                "experiment-old",
                "network",
                "obsvar",
                "layer",
                "model",
                "modvar",
                "time",
            ],
            None,
            "./project/experiment/map/",
        ),
        (
            "get_map",
            [
                "project",
                "experiment",
                "network",
                "obsvar",
                "layer",
                "model",
                "modvar",
                "time",
            ],
            None,
            "./project/experiment/map/with_time",
        ),
        (
            "get_scatter",
            [
                "project",
                "experiment",
                "network",
                "obsvar",
                "layer",
                "model",
                "modvar",
                "time",
            ],
            None,
            "./project/experiment/scat/time",
        ),
        (
            "get_scatter",
            [
                "project",
                "experiment-old",
                "network",
                "obsvar",
                "layer",
                "model",
                "modvar",
                "test",
            ],
            None,
            "./project/experiment/scat/",
        ),
        (
            "get_profiles",
            ["project", "experiment", "region", "network", "obsvar"],
            None,
            "./project/experiment/profiles/",
        ),
        (
            "get_heatmap_timeseries",
            ["project", "experiment", "region", "network", "obsvar", "layer"],
            None,
            "./project/experiment/hm/ts/region-network-obsvar-layer",
        ),
        (
            "get_heatmap_timeseries",
            ["project", "experiment-old", "region", "network", "obsvar", "layer"],
            None,
            "project/experiment/hm/ts/stats_ts.json",
        ),
        # TODO: Missing test case for heatmap_ts with the middle version format.
        (
            "get_forecast",
            ["project", "experiment", "region", "network", "obsvar", "layer"],
            None,
            "./project/experiment/forecast/",
        ),
        (
            "get_fairmode",
            ["project", "experiment", "region", "network", "obsvar", "layer"],
            None,
            "./project/experiment/fairmode/",
        ),
        (
            "get_gridded_map",
            ["project", "experiment", "obsvar", "model"],
            None,
            "./project/experiment/contour/",
        ),
        (
            "get_report",
            ["project", "experiment", "title"],
            None,
            "./reports/project/experiment/",
        ),
    ),
)

PUT_PARAMETRIZATION = pytest.mark.parametrize(
    "fun,args,kwargs",
    (
        ("glob_stats", ["project", "experiment", "frequency"], None),
        (
            "timeseries",
            ["project", "experiment", "location", "network", "obsvar", "layer"],
            None,
        ),
        (
            "contour",
            ["project", "experiment", "obsvar", "model"],
            {"timestep": "timestep"},
        ),
        (
            "timeseries_weekly",
            ["project", "experiment", "location", "network", "obsvar", "layer"],
            None,
        ),
        ("config", ["project", "experiment"], None),
        ("menu", ["project", "experiment"], None),
        ("statistics", ["project", "experiment"], None),
        ("ranges", ["project", "experiment"], None),
        ("regions", ["project", "experiment"], None),
        ("models_style", ["project"], None),
        ("models_style", ["project"], {"experiment": "experiment"}),
        (
            "map",
            [
                "project",
                "experiment",
                "network",
                "obsvar",
                "layer",
                "model",
                "modvar",
                "time",
            ],
            None,
        ),
        (
            "map",
            [
                "project",
                "experiment",
                "network",
                "obsvar",
                "layer",
                "model",
                "modvar",
                "time",
            ],
            None,
        ),
        (
            "scatter",
            [
                "project",
                "experiment",
                "network",
                "obsvar",
                "layer",
                "model",
                "modvar",
                "time",
            ],
            None,
        ),
        (
            "scatter",
            [
                "project",
                "experiment",
                "network",
                "obsvar",
                "layer",
                "model",
                "modvar",
                "time",
            ],
            None,
        ),
        ("profiles", ["project", "experiment", "station", "network", "obsvar"], None),
        (
            "heatmap_timeseries",
            ["project", "experiment", "region", "network", "obsvar", "layer"],
            None,
        ),
        (
            "forecast",
            ["project", "experiment", "station", "network", "obsvar", "layer"],
            None,
        ),
        (
            "fairmode",
            ["project", "experiment", "station", "network", "obsvar", "layer"],
            None,
        ),
        ("gridded_map", ["project", "experiment", "obsvar", "model"], None),
        ("report", ["project", "experiment", "title"], None),
    ),
)


@pytest.mark.asyncio
@TESTDB_PARAMETRIZATION
@GET_PARAMETRIZATION
async def test_getter(testdb: str, fun: str, args: list, kwargs: dict, expected):
    """
    This test tests that data is read as expected from a static, fixed database.
    """
    with aerovaldb.open(testdb) as db:
        f = getattr(db, fun)

        if kwargs is not None:
            data = await f(*args, **kwargs)
        else:
            data = await f(*args)

        assert data["path"] == expected


@TESTDB_PARAMETRIZATION
@GET_PARAMETRIZATION
def test_getter_sync(testdb: str, fun: str, args: list, kwargs: dict, expected):
    with aerovaldb.open(testdb) as db:
        f = getattr(db, fun)

        if kwargs is not None:
            data = f(*args, **kwargs)
        else:
            data = f(*args)

        assert data["path"] == expected


@TESTDB_PARAMETRIZATION
@GET_PARAMETRIZATION
def test_getter_json_str(testdb: str, fun: str, args: list, kwargs: dict, expected):
    with aerovaldb.open(testdb) as db:
        f = getattr(db, fun)

        if kwargs is not None:
            data = f(*args, access_type=aerovaldb.AccessType.JSON_STR, **kwargs)
        else:
            data = f(*args, access_type=aerovaldb.AccessType.JSON_STR)

        data = simplejson.loads(data)
        assert data["path"] == expected


@pytest.mark.asyncio
@IMPLEMENTATION_PARAMETRIZATION
@PUT_PARAMETRIZATION
async def test_setters(dbtype: str, fun: str, args: list, kwargs: dict, tmpdb):
    """
    This test tests that you read back the expected data, once you have written
    to a fresh db, assuming the same arguments.
    """
    with tmpdb as db:
        get = getattr(db, f"get_{fun}")
        put = getattr(db, f"put_{fun}")

        expected = fun + str(random.randint(0, 100000))
        if kwargs is not None:
            await put({"data": expected}, *args, **kwargs)

            data = await get(*args, **kwargs)
        else:
            await put({"data": expected}, *args)

            data = await get(*args)

        assert data["data"] == expected


@IMPLEMENTATION_PARAMETRIZATION
@PUT_PARAMETRIZATION
def test_setters_sync(fun: str, args: list, kwargs: dict, tmpdb):
    """
    This test tests that you read back the expected data, once you have written
    to a fresh db, assuming the same arguments.
    """
    with tmpdb as db:
        get = getattr(db, f"get_{fun}")
        put = getattr(db, f"put_{fun}")

        expected = fun + str(random.randint(0, 100000))
        if kwargs is not None:
            put({"data": expected}, *args, **kwargs)

            data = get(*args, **kwargs)
        else:
            put({"data": expected}, *args)

            data = get(*args)

        assert data["data"] == expected


@IMPLEMENTATION_PARAMETRIZATION
@PUT_PARAMETRIZATION
def test_setters_json_str(fun: str, args: list, kwargs: dict, tmpdb):
    with tmpdb as db:
        get = getattr(db, f"get_{fun}")
        put = getattr(db, f"put_{fun}")

        expected = fun + str(random.randint(0, 100000))
        if kwargs is not None:
            put(aerovaldb.utils.json_dumps_wrapper({"data": expected}), *args, **kwargs)

            data = get(*args, **kwargs)
        else:
            put(aerovaldb.utils.json_dumps_wrapper({"data": expected}), *args)

            data = get(*args)

        assert data["data"] == expected


@IMPLEMENTATION_PARAMETRIZATION
def test_write_and_read_of_nan(tmpdb):
    with tmpdb as db:
        data = dict(value=float("nan"))

        db.put_by_uri(data, "/v0/experiments/project")

        read = db.get_by_uri("/v0/experiments/project")

        # See Additional Notes on #59
        # https://github.com/metno/aerovaldb/pull/59
        assert read["value"] is None


@pytest.mark.asyncio
@TESTDB_PARAMETRIZATION
async def test_file_does_not_exist(testdb):
    with aerovaldb.open(testdb) as db:
        with pytest.raises(FileNotFoundError):
            await db.get_config(
                "non-existent-project",
                "experiment",
            )


@TESTDB_PARAMETRIZATION
def test_getter_with_default(testdb):
    with aerovaldb.open(testdb) as db:
        data = db.get_by_uri(
            "/v0/experiments/non-existent-project", default={"data": "test"}
        )

        assert data["data"] == "test"


@TESTDB_PARAMETRIZATION
def test_getter_with_default_error(testdb):
    with aerovaldb.open(testdb) as db:
        with pytest.raises(simplejson.JSONDecodeError):
            db.get_by_uri(
                "/v0/report/project/experiment/invalid-json",
                default={"data": "data"},
            )


@TESTDB_PARAMETRIZATION
def test_version1(testdb):
    """ """
    with aerovaldb.open(testdb) as db:
        assert str(db._get_version("project", "experiment")) == "0.13.5"


@TESTDB_PARAMETRIZATION
def test_version2(testdb):
    """ """
    with aerovaldb.open(testdb) as db:
        assert str(db._get_version("project", "experiment-old")) == "0.0.5"


@pytest.mark.dependency(scope="session", name="test_list_all")
@TESTDB_PARAMETRIZATION
def test_list_all(testdb):
    with aerovaldb.open(testdb) as db:
        assert len(db.list_all()) == 50


@TESTDB_PARAMETRIZATION
def test_list_timeseries(testdb):
    with aerovaldb.open(testdb) as db:
        timeseries = db.list_timeseries("project", "experiment")

        assert len(list(timeseries)) == 1


@IMPLEMENTATION_PARAMETRIZATION
def test_rm_experiment_data(tmpdb):
    with aerovaldb.open("json_files:./tests/test-db/json") as db:
        copy_db_contents(db, tmpdb)
        assert len(list(db.list_all())) == len(list(tmpdb.list_all()))

        tmpdb.rm_experiment_data("project", "experiment")

        assert len(list(tmpdb.list_all())) == 30


@TESTDB_PARAMETRIZATION
@pytest.mark.parametrize(
    "sub_path",
    (
        pytest.param("img/pixel.avif"),
        pytest.param("img/pixel.gif"),
        pytest.param("img/pixel.jpeg"),
        pytest.param("img/pixel.jpg"),
        pytest.param("img/pixel.png"),
        pytest.param("img/pixel.webp"),
    ),
)
def test_get_report_image(testdb, sub_path: str):
    with aerovaldb.open(testdb) as db:
        blob = db.get_report_image(
            "project",
            "experiment",
            sub_path,
            access_type=aerovaldb.AccessType.BLOB,
        )
        assert isinstance(blob, bytes)
        assert len(blob) > 0


@IMPLEMENTATION_PARAMETRIZATION
def test_put_report_image(tmpdb):
    with open("tests/test-db/json/reports/project/experiment/img/pixel.png", "rb") as f:
        data = f.read()

    with tmpdb as db:
        db.put_report_image(data, "project", "experiment", "pixel.png")

        blob = db.get_report_image("project", "experiment", "pixel.png")

    assert isinstance(blob, bytes)
    assert len(blob) > 0


@IMPLEMENTATION_PARAMETRIZATION
def test_serialize_set(tmpdb):
    with tmpdb as db:
        db.put_config({"set": {"a", "b", "c"}}, "test", "test")


@TESTDB_PARAMETRIZATION
def test_get_times(testdb):
    with aerovaldb.open(testdb) as db:
        for uri in db.list_all():
            mtime = db.get_by_uri(uri, access_type="MTIME")
            ctime = db.get_by_uri(uri, access_type="CTIME")

            assert isinstance(mtime, datetime.datetime)
            assert isinstance(ctime, datetime.datetime)
            assert mtime.year >= 2024 and mtime < datetime.datetime.now()
            assert ctime.year >= 2024 and ctime < datetime.datetime.now()


@TESTDB_PARAMETRIZATION
def test_get_experiment_mtime(testdb):
    with aerovaldb.open(testdb) as db:
        for exp in ["experiment", "experiment-old"]:
            mtime = db.get_experiment_mtime("project", exp)

            assert isinstance(mtime, datetime.datetime)
            assert mtime.year >= 2024 and mtime < datetime.datetime.now()


TEST_IMAGES = {
    ".webp": pathlib.Path("tests/test_img/test.webp"),
    ".png": pathlib.Path("tests/test_img/test.png"),
}


@TESTDB_PARAMETRIZATION
def test_get_map_overlay(testdb):
    with aerovaldb.open(testdb) as db:
        data: bytes = db.get_map_overlay(
            "project",
            "experiment",
            "source",
            "variable",
            "date",
            access_type=aerovaldb.AccessType.BLOB,
        )

        assert filetype.guess_extension(data) == "png"


@IMPLEMENTATION_PARAMETRIZATION
@pytest.mark.parametrize(
    "expected_extension",
    (
        pytest.param(".png", id="png"),
        pytest.param(".webp", id="webp"),
    ),
)
def test_put_map_overlay(tmpdb, expected_extension: str):
    with open(TEST_IMAGES[expected_extension], "rb") as f:
        input_data = f.read()

    with tmpdb as db:
        db.put_map_overlay(
            input_data,
            "project",
            "experiment",
            "source",
            "variable",
            "date",
        )

        data: bytes = db.get_map_overlay(
            "project",
            "experiment",
            "source",
            "variable",
            "date",
            access_type=aerovaldb.AccessType.BLOB,
        )
        assert data == input_data

        if isinstance(tmpdb, aerovaldb.jsondb.AerovalJsonFileDB):
            file_path: str = db.get_map_overlay(
                "project",
                "experiment",
                "source",
                "variable",
                "date",
                access_type=aerovaldb.AccessType.FILE_PATH,
            )
            assert file_path.endswith(expected_extension)


@TESTDB_PARAMETRIZATION
def test_get_map_filtering(testdb):
    with aerovaldb.open(testdb) as db:
        data = db.get_map(
            "project",
            "experiment",
            "network",
            "obsvar",
            "layer",
            "model",
            "modvar",
            "time2",
            frequency="frequency",
            season="season",
        )

    assert "frequency" in data[0]
    assert not "excluded_frequency" in data[0]
    assert "season" in data[0]["frequency"]
    assert not "excluded_season" in data[0]["frequency"]


@TESTDB_PARAMETRIZATION
def test_list_glob_stats(testdb):
    with aerovaldb.open(testdb) as db:
        glob_stats = db.list_glob_stats("project", "experiment")

        assert len(glob_stats) == 1



import argparse
import logging

from .. import AccessType, AerovalDB, open
from ..utils import async_and_sync

logger = logging.getLogger(__name__)


@async_and_sync
async def copy_db_contents(source: str | AerovalDB, dest: str | AerovalDB):
    """
    Utility function for copying the contents of one db connection to another
    Currently this implementation requires the destination db to be empty.

    :param source : Instance of AerovalDB or resource string passed to
        aerovaldb.open()
    :param dest : Instance of AerovalDB or resource string passed to
        aerovaldb.open()

    :raises : ValueError
        If destination is not empty.
    :raises : IOError
        Number of items in destination is not the same as the source after copy.
    """
    if isinstance(source, str):
        source = open(source)

    if isinstance(dest, str):
        dest = open(dest)

    if len(await dest.list_all()) > 0:
        ValueError("Destination database is not empty.")

    for i, uri in enumerate(await source.list_all()):
        uri = str(uri)
        logger.info(f"Processing item {i} of {len(await source.list_all())}")
        access = AccessType.JSON_STR
        if uri.startswith("/v0/report-image/") or uri.startswith("/v0/map-overlay/"):
            access = AccessType.BLOB
        data = await source.get_by_uri(uri, access_type=access)

        await dest.put_by_uri(data, uri)

    dst_len = len(await dest.list_all())
    src_len = len(await source.list_all())
    if dst_len != src_len:
        raise IOError(
            f"Unexpected number of items in destination after copy. Expected {src_len}, got {dst_len}"
        )


def main():
    parser = argparse.ArgumentParser(
        description="Utility for copying contents from one aerovaldb resource to another."
    )
    parser.add_argument(
        "source", type=str, help="Resource string of the source AerovalDB instance"
    )
    parser.add_argument(
        "dest", type=str, help="Resource string of the destination AerovalDB instance"
    )

    args = parser.parse_args()

    source = args.source
    dest = args.dest
    copy_db_contents(source, dest)


if __name__ == "__main__":
    main()



import sys

if sys.version_info >= (3, 12):
    from typing import override
else:
    from typing_extensions import override

from ..routes import Route


class QueryEntry:
    """Class representing information about an asset. Information
    is represented as a string representing the URI of the asset
    but additional information is available, namely type (See
    AssetType) and meta (dict containing metadata about the asset).
    """

    def __init__(self, uri: str, type: Route, meta: dict[str, str]):
        self._uri = uri
        self._type = type
        self._meta = meta

    @property
    def uri(self) -> str:
        """The URI of the asset."""
        return self._uri

    @property
    def type(self) -> Route:
        """The type of the asset this represents (see Route enum for details)."""
        return self._type

    @property
    def meta(self) -> dict[str, str]:
        """Dictionary containing metadata about the asset. Available keys should
        reflect the args passed to put for this assettype.

        :return: Dict containing metadata.
        """
        return self._meta

    @override
    def __str__(self) -> str:
        return self.uri



import asyncio
import functools
from typing import Callable, ParamSpec, TypeVar

# Workaround to ensure function signature of the decorated function is shown correctly
# Solution from here: https://stackoverflow.com/questions/74074580/how-to-avoid-losing-type-hinting-of-decorated-function
P = ParamSpec("P")
T = TypeVar("T")


def has_async_loop():
    is_async = False
    try:
        loop = asyncio.get_running_loop()
        if loop is not None:
            is_async = True
    except RuntimeError:
        is_async = False
    return is_async


def async_and_sync(function: Callable[P, T]) -> Callable[P, T]:
    """Wrap an async method to a sync method.

    This allows to run the async method in both async and sync contexts transparently
    without any additional code.

    :args function: function/property to wrap
    :return: modified function
    """

    @functools.wraps(function)
    def async_and_sync_wrap(*args, **kwargs):
        if has_async_loop():
            return function(*args, **kwargs)
        else:
            return asyncio.run(function(*args, **kwargs))

    return async_and_sync_wrap


@async_and_sync
async def run_until_finished(coroutine):
    """
    Takes a aio coroutine, runs it and waits for it to finish.

    :param coroutine : The coroutine to be ran.
    :return
        The result from running coroutine.
    """
    task = asyncio.create_task(coroutine)
    await asyncio.wait(task)
    return task.result()



import logging
import os
import sys
from abc import ABC, abstractmethod
from collections import defaultdict, deque
from pathlib import Path
from typing import Hashable, TypedDict

if sys.version_info >= (3, 12):
    from typing import override
else:
    from typing_extensions import override
logger = logging.getLogger(__name__)


class CacheMissError(FileNotFoundError):
    """
    Raised by cache implementations when a cache miss occurs
    that the implementation is unable to handle (by eg.
    delegating it).
    """

    pass


class LRUQueue:
    """
    Small helper class that efficiently maintains a LRUQueue
    by combining a set and deque to maintain a unique constraint
    on the queue. Re-adding an element will return it to the
    end of the queue.
    """

    def __init__(self):
        self._set = set()
        self._deque = deque()

    @property
    def size(self) -> int:
        """Returns the lenth of the queue."""
        return len(self._set)

    def add(self, item: Hashable):
        """
        Adds an item to the queue.
        """
        if item in self._set:
            self._deque.remove(item)

        self._set.add(item)
        self._deque.appendleft(item)

    def pop(self) -> Hashable:
        """Removes and returns the top item from the queue."""
        item = self._deque.pop()
        self._set.remove(item)
        return item

    def remove(self, item: Hashable):
        """
        Removes an item from the queue.
        """
        if item in self._set:
            self._set.remove(item)
            self._deque.remove(item)


class BaseCache(ABC):
    def __init__(self):
        pass

    @property
    @abstractmethod
    def hit_count(self) -> int:
        raise NotImplementedError

    @property
    @abstractmethod
    def size(self) -> int:
        raise NotImplementedError

    @property
    @abstractmethod
    def miss_count(self) -> int:
        raise NotImplementedError

    @abstractmethod
    def get(self, key: str, *, bypass_cache: bool = False) -> str:
        raise NotImplementedError

    @abstractmethod
    def put(self, obj: str, *, key: str):
        raise NotImplementedError

    @abstractmethod
    def clear(self) -> None:
        raise NotImplementedError

    @abstractmethod
    def evict(self, file_path: str) -> None:
        raise NotImplementedError

    @abstractmethod
    def is_valid(self, key: str) -> bool:
        raise NotImplementedError


class CacheEntry(TypedDict):
    file: str

    last_modified: float


class LRUFileCache(BaseCache):
    """
    Implements an in-memory LRU cache for file content.
    """

    def __init__(self, *, max_size: int):
        """
        :param max_size : The maximum size of the cache in terms of number of entries / files.

        Files will be ejected based on least recently used, when full.
        """
        self._hit_count: int = 0
        self._miss_count: int = 0
        self._max_size: int = max_size

        # Stores the actual cached content, indexed by canonical file path.
        self._entries: defaultdict[str, CacheEntry | None] = defaultdict(lambda: None)

        # Stores queue of cache accesses, used for implementing LRU logic.
        self._queue = LRUQueue()

    def _get_entry(self, abspath: str):
        """Returns an element from the cache."""
        self._queue.add(abspath)
        self._hit_count = self._hit_count + 1
        logger.debug(f"Returning contents from file {abspath} from cache.")
        return self._entries[abspath]["file"]  # type: ignore

    @override
    def clear(self) -> None:
        logger.debug("Cache cleared.")

        self._entries = defaultdict(lambda: None)
        self._queue = LRUQueue()

        # Tally of cache hits and misses.
        self._hit_count = 0
        self._miss_count = 0

    @property
    @override
    def hit_count(self) -> int:
        """Returns the number of cache hits since the last `invalidate_all()` call.

        Note:
        -----
        This does not include calls with `no_cache=True`
        """
        return self._hit_count

    @property
    @override
    def size(self) -> int:
        """Returns the current size of the cache in terms of number of elements."""
        return self._queue.size

    @property
    @override
    def miss_count(self) -> int:
        """Returns the number of cache misses since the last `invalidate_all()` call.

        Note:
        -----
        This does not include calls with `no_cache=True`
        """
        return self._miss_count

    def _canonical_file_path(self, file_path: str | Path) -> str:
        """
        Returns an absolute file path with symlinks removed for a file to
        ensure correct lookup.

        :param file_path : The file path.

        :return : The file path converted to canonical file path.
        """
        return str(os.path.realpath(file_path))

    def _read_file(self, file_path: str | Path) -> str:
        abspath = self._canonical_file_path(file_path)
        logger.debug(f"Reading file {abspath}")

        with open(abspath, "r") as f:
            return f.read()

    def _put_entry(self, abspath: str, *, obj: str):
        self._entries[abspath] = {
            "file": obj,
            "last_modified": os.path.getmtime(abspath),
        }
        while self.size > self._max_size:
            key = self._queue.pop()
            self.evict(str(key))

    @override
    def get(self, key: str, *, bypass_cache: bool = False) -> str:
        abspath = self._canonical_file_path(key)

        if bypass_cache:
            return self._read_file(abspath)

        if self.is_valid(abspath):
            return self._get_entry(abspath)

        self._miss_count += 1
        obj = self._read_file(abspath)
        self._queue.add(abspath)
        self._put_entry(abspath, obj=obj)
        return obj

    @override
    def put(self, obj, *, key: str):
        abspath = self._canonical_file_path(key)
        self._put_entry(abspath, obj=obj)
        logger.debug("%s of %s entries in cache", self.size, self._max_size)

    @override
    def evict(self, file_path: str | Path) -> None:
        """
        Invalidates the cache for a file path, ensuring it will be re-read on the next read.

        :param file_path : The file path to invalidate cache for.
        """
        abspath = self._canonical_file_path(file_path)
        logger.debug(f"Clearing cache for file {abspath}.")
        if abspath in self._entries:
            del self._entries[abspath]
            self._queue.remove(abspath)

    @override
    def is_valid(self, file_path: str | Path) -> bool:
        """
        Checks whether a cache element is valid.

        :param file_path: The file path to check for.

        :returns : Boolean indicating cache validity.
        """
        abspath = self._canonical_file_path(file_path)

        cache = self._entries[abspath]
        if cache is None:
            return False

        if not os.path.exists(abspath):
            return False

        if os.path.getmtime(abspath) > cache["last_modified"]:
            return False

        return True


class KeyCacheDecorator(BaseCache):
    """Decorator for other cache implementations which extends it with support for
    sub-parts of a json file.
    """

    def __init__(self, cache: BaseCache, *, max_size: int = 64):
        if not isinstance(cache, BaseCache):
            raise TypeError(f"Cache is of type {type(cache)}, expected BaseCache")

        self._cache = cache

        self._entries: dict[str, CacheEntry | None] = defaultdict(lambda: None)
        self._queue = LRUQueue()

        self._miss_count = 0
        self._hit_count = 0
        self._max_size = max_size

    def _split_key(self, key: str) -> tuple[str, str | None]:
        splt = key.split("::")
        if len(splt) == 1:
            return (splt[0], None)
        elif len(splt) == 2:
            return tuple(splt)  # type: ignore

        raise ValueError(
            f"Unexpected number of elements in '{key}'. Expected 1 or 2, got {len(splt)}."
        )

    @property
    @override
    def hit_count(self) -> int:
        return self._hit_count

    @property
    @override
    def size(self) -> int:
        return self._queue.size

    @property
    @override
    def miss_count(self) -> int:
        return self._miss_count

    @override
    def get(self, key: str, *, bypass_cache: bool = False) -> str:
        fp, k = self._split_key(key)

        if k is None:
            return self._cache.get(fp, bypass_cache=bypass_cache)

        if self.is_valid(key):
            if (entry := self._entries[key]) is not None:
                self._hit_count += 1
                return entry["file"]

        self._miss_count += 1
        raise CacheMissError

    @override
    def put(self, obj, *, key: str) -> None:
        fp, _ = self._split_key(key)
        self._entries[key] = {
            "file": obj,
            "last_modified": os.path.getmtime(fp),
        }
        self._queue.add(key)
        while self.size > self._max_size:
            key = self._queue.pop()  # type: ignore
            self.evict(str(key))

        logger.debug("%s of %s entries in cache", self.size, self._max_size)

    @override
    def clear(self) -> None:
        self._entries = defaultdict(lambda: None)
        self._miss_count = 0
        self._hit_count = 0
        self._queue = LRUQueue()

    @override
    def evict(self, key: str) -> None:
        logger.debug(f"Invalidating cache for key {key}.")
        if key in self._entries:
            del self._entries[key]
            self._queue.remove(key)

    @override
    def is_valid(self, key: str) -> bool:
        fp, k = self._split_key(key)
        if k is None:
            # File access is delegated to sub-cache.
            return self._cache.is_valid(fp)

        cache = self._entries[key]
        if cache is None:
            return False

        if not os.path.exists(fp):
            return False

        if os.path.getmtime(fp) > cache["last_modified"]:
            return False

        return True



import functools
import os
import sys
import warnings

if sys.version_info >= (3, 10):
    from importlib.metadata import EntryPoints, entry_points
else:
    from importlib_metadata import EntryPoints, entry_points

from .aerovaldb import AerovalDB


def _build_db_engines(entrypoints: EntryPoints) -> dict[str, AerovalDB]:
    backend_entrypoints: dict[str, AerovalDB] = {}
    for entrypoint in entrypoints:
        name = entrypoint.name
        if name in backend_entrypoints:
            warnings.warn(
                f"found multiple versions of {entrypoint.group} entrypoint {name} for {entrypoint.value}"
            )
            continue
        try:
            backend_entrypoints[name] = entrypoint.load()
        except Exception as ex:
            warnings.warn(f"Engine {name!r} loading failed:\n{ex}", RuntimeWarning)
    return backend_entrypoints


@functools.lru_cache(maxsize=1)
def list_engines() -> dict[str, AerovalDB]:
    """
    Return a dictionary of available AerovalDB classes.

    Returns
    -------
    dictionary

    Notes
    -----
    This function lives in the backends namespace (``dbs = aerovaldb.list_engines()``).

    """
    entrypoints = entry_points(group="aerovaldb")
    return _build_db_engines(entrypoints)


def open(resource) -> AerovalDB:
    """Open an AerovalDB instance, sending args and kwargs
    directly to the `AervoalDB()` function

    :param resource: the resource-name for the database. The resource can be
        - 'entrypoint:path', with path being the location where the database should be generated
        (eg. 'json_files:.')
        - 'path', being a json_files dabasase (for example, '.' is equivalent to 'json_files:.')
        - ':memory:' an sqlite in-memory database. Contents are not persistently
        stored!
    :return: an implementation-instance of AerovalDB openend to a location

    Examples

    >>> import aerovaldb
    >>> with aerovaldb.open(":memory:") as db:
    ...     db.put_experiments({'test': 'test'}, 'project')
    ...     db.get_experiments('project')
    {'test': 'test'}

    >>> import aerovaldb
    >>> db = aerovaldb.open(":memory:")
    >>> db.put_experiments({'test': 'test'}, 'project')
    >>> db.get_experiments('project')
    {'test': 'test'}
    """
    if resource == ":memory:":
        # Special case for sqlite in memory database.
        name = "sqlitedb"
        path = ":memory:"

    elif ":" in resource:
        parts = resource.split(":")
        if len(parts) > 1:
            name = parts[0]
            path = ":".join(parts[1:])
        else:
            # TODO check if path contains a aerovaldb cfg file
            name = "json_files"
            path = resource
    else:
        fileextension = os.path.splitext(resource)[1]
        if fileextension in [".db", ".sqlite"]:
            # Sqlite file.
            name = "sqlitedb"
            path = resource
        else:
            # Assume directory and json.
            name = "json_files"
            path = resource

    aerodb = list_engines()[name]

    return aerodb(path)  # type: ignore



import datetime
import importlib.metadata
import inspect
import logging
import os
import sqlite3
import sys
from hashlib import md5
from typing import Any, Awaitable, Callable

import simplejson  # type: ignore
from async_lru import alru_cache
from packaging.version import Version

import aerovaldb
from aerovaldb.utils.filter import (
    filter_contour,
    filter_heatmap,
    filter_map,
    filter_regional_stats,
)
from aerovaldb.utils.query import QueryEntry
from aerovaldb.utils.string_mapper import (
    PriorityMapper,
    StringMapper,
    VersionConstraintMapper,
)

from ..aerovaldb import AerovalDB
from ..exceptions import UnsupportedOperation, UnusedArguments
from ..lock import FakeLock, FileLock
from ..routes import *
from ..types import AccessType
from ..utils import (
    async_and_sync,
    build_uri,
    extract_substitutions,
    json_dumps_wrapper,
    parse_uri,
)

if sys.version_info >= (3, 12):
    from typing import override
else:
    from typing_extensions import override

logger = logging.getLogger(__name__)


def _column_titles_from_route(route: Route):
    return extract_substitutions(route.value)


class AerovalSqliteDB(AerovalDB):
    """
    Allows reading and writing from sqlite3 database files.
    """

    SQLITE_TIMESTAMP_FORMAT = "%Y-%m-%d %H:%M:%S"

    TABLE_COLUMN_NAMES = {
        "glob_stats": _column_titles_from_route(Route.GLOB_STATS),
        "contour": _column_titles_from_route(Route.CONTOUR),
        "contour1": ["project", "experiment", "obsvar", "model", "timestep"],
        "timeseries": _column_titles_from_route(Route.TIMESERIES),
        "timeseries_weekly": _column_titles_from_route(Route.TIMESERIES_WEEKLY),
        "experiments": _column_titles_from_route(Route.EXPERIMENTS),
        "config": _column_titles_from_route(Route.CONFIG),
        "menu": _column_titles_from_route(Route.MENU),
        "statistics": _column_titles_from_route(Route.STATISTICS),
        "ranges": _column_titles_from_route(Route.RANGES),
        "regions": _column_titles_from_route(Route.REGIONS),
        "models_style0": ["project", "experiment"],
        "models_style1": ["project"],
        "map0": [
            "project",
            "experiment",
            "network",
            "obsvar",
            "layer",
            "model",
            "modvar",
            "time",
        ],
        "map1": [
            "project",
            "experiment",
            "network",
            "obsvar",
            "layer",
            "model",
            "modvar",
        ],
        "scatter0": [
            "project",
            "experiment",
            "network",
            "obsvar",
            "layer",
            "model",
            "modvar",
            "time",
        ],
        "scatter1": [
            "project",
            "experiment",
            "network",
            "obsvar",
            "layer",
            "model",
            "modvar",
        ],
        "profiles": _column_titles_from_route(Route.PROFILES),
        "heatmap_timeseries0": [
            "project",
            "experiment",
            "region",
            "network",
            "obsvar",
            "layer",
        ],
        "heatmap_timeseries1": ["project", "experiment", "network", "obsvar", "layer"],
        "heatmap_timeseries2": [
            "project",
            "experiment",
        ],
        "forecast": _column_titles_from_route(Route.FORECAST),
        "fairmode": _column_titles_from_route(Route.FAIRMODE),
        "gridded_map": _column_titles_from_route(Route.GRIDDED_MAP),
        "report": _column_titles_from_route(Route.REPORT),
        "reportimages": _column_titles_from_route(Route.REPORT_IMAGE),
        "mapoverlay": _column_titles_from_route(Route.MAP_OVERLAY),
    }

    TABLE_NAME_TO_ROUTE = {
        "glob_stats": Route.GLOB_STATS,
        "contour": Route.CONTOUR,
        "contour1": Route.CONTOUR_TIMESPLIT,
        "timeseries": Route.TIMESERIES,
        "timeseries_weekly": Route.TIMESERIES_WEEKLY,
        "experiments": Route.EXPERIMENTS,
        "config": Route.CONFIG,
        "menu": Route.MENU,
        "statistics": Route.STATISTICS,
        "ranges": Route.RANGES,
        "regions": Route.REGIONS,
        "models_style0": Route.MODELS_STYLE,
        "models_style1": Route.MODELS_STYLE,
        "map0": Route.MAP,
        "map1": Route.MAP,
        "scatter0": Route.SCATTER,
        "scatter1": Route.SCATTER,
        "profiles": Route.PROFILES,
        "heatmap_timeseries0": Route.HEATMAP_TIMESERIES,
        "heatmap_timeseries1": Route.HEATMAP_TIMESERIES,
        "heatmap_timeseries2": Route.HEATMAP_TIMESERIES,
        "forecast": Route.FORECAST,
        "fairmode": Route.FAIRMODE,
        "gridded_map": Route.GRIDDED_MAP,
        "report": Route.REPORT,
        "reportimages": Route.REPORT_IMAGE,
        "mapoverlay": Route.MAP_OVERLAY,
    }

    def __init__(self, database: str, /, **kwargs):
        use_locking = os.environ.get("AVDB_USE_LOCKING", "")
        if use_locking == "0" or use_locking == "":
            self._use_real_lock = False
        else:
            self._use_real_lock = True

        self._dbfile = database

        if not os.path.exists(database):
            self._con = sqlite3.connect(database)
            self._initialize_db()
        else:
            self._con = sqlite3.connect(database)
            if not self._get_metadata_by_key("created_by") == "aerovaldb":
                ValueError(f"Database {database} is not a valid aerovaldb database.")

        self._con.row_factory = sqlite3.Row

        self.TABLE_NAME_LOOKUP = StringMapper(
            {
                Route.GLOB_STATS: "glob_stats",
                Route.REGIONAL_STATS: "glob_stats",
                Route.HEATMAP: "glob_stats",
                Route.CONTOUR: "contour",
                Route.CONTOUR_TIMESPLIT: "contour1",
                Route.TIMESERIES: "timeseries",
                Route.TIMESERIES_WEEKLY: "timeseries_weekly",
                Route.EXPERIMENTS: "experiments",
                Route.CONFIG: "config",
                Route.MENU: "menu",
                Route.STATISTICS: "statistics",
                Route.RANGES: "ranges",
                Route.REGIONS: "regions",
                Route.MODELS_STYLE: PriorityMapper(
                    {
                        "models_style0": "{project}/{experiment}",
                        "models_style1": "{project}",
                    }
                ),
                Route.MAP: [
                    VersionConstraintMapper(
                        "map0",
                        min_version="0.13.2",
                    ),
                    VersionConstraintMapper(
                        "map1",
                        max_version="0.13.2",
                    ),
                ],
                Route.SCATTER: [
                    VersionConstraintMapper(
                        "scatter0",
                        min_version="0.13.2",
                    ),
                    VersionConstraintMapper(
                        "scatter1",
                        max_version="0.13.2",
                    ),
                ],
                Route.PROFILES: "profiles",
                Route.HEATMAP_TIMESERIES: [
                    VersionConstraintMapper(
                        "heatmap_timeseries0",
                        min_version="0.13.2",  # https://github.com/metno/pyaerocom/blob/4478b4eafb96f0ca9fd722be378c9711ae10c1f6/setup.cfg
                    ),
                    VersionConstraintMapper(
                        "heatmap_timeseries1",
                        min_version="0.12.2",
                        max_version="0.13.2",
                    ),
                    VersionConstraintMapper(
                        "heatmap_timeseries2",
                        max_version="0.12.2",
                    ),
                ],
                Route.FORECAST: "forecast",
                Route.FAIRMODE: "fairmode",
                Route.GRIDDED_MAP: "gridded_map",
                Route.REPORT: "report",
                Route.REPORT_IMAGE: "reportimages",
                Route.MAP_OVERLAY: "mapoverlay",
            },
            version_provider=self._get_version,
        )

        self.FILTERS: dict[Route, Callable[..., Awaitable[Any]]] = {
            Route.REGIONAL_STATS: filter_regional_stats,
            Route.HEATMAP: filter_heatmap,
            Route.CONTOUR: filter_contour,
            Route.MAP: filter_map,
        }

    @async_and_sync
    @alru_cache(maxsize=2048)
    async def _get_version(self, project: str, experiment: str) -> Version:
        """
        Returns the version of pyaerocom used to generate the files for a given project
        and experiment.

        :param project : Project ID.
        :param experiment : Experiment ID.

        :return : A Version object.
        """
        try:
            config = await self.get_config(project, experiment)
        except FileNotFoundError:
            try:
                # If pyaerocom is installed in the current environment, but no config has
                # been written, we use the version of the installed pyaerocom. This is
                # important for tests to work correctly, and for files to be written
                # correctly if the config file happens to be written after data files.
                version = Version(importlib.metadata.version("pyaerocom"))
            except importlib.metadata.PackageNotFoundError:
                version = Version("0.0.1")
            finally:
                return version

        try:
            version_str = config["exp_info"]["pyaerocom_version"]
            version = Version(version_str)
        except KeyError:
            version = Version("0.0.1")

        return version

    def _get_metadata_by_key(self, key: str) -> str:
        """
        Returns the value associated with a key from the metadata
        table.
        """
        cur = self._con.cursor()

        cur.execute(
            """
            SELECT value FROM metadata
            WHERE key = ?
            """,
            (key,),
        )
        return cur.fetchone()[0]

    def _set_metadata_by_key(self, key: str, value: str):
        """ """
        cur = self._con.cursor()

        cur.execute(
            """
            INSERT OR REPLACE INTO metadata(key, value)
            VALUES(?, ?)
            """,
            (key, value),
        )

    def _initialize_db(self):
        """Given an existing sqlite connection or sqlite3 database
        identifier string, initializes the database so it has the
        necessary tables.
        """
        cur = self._con.cursor()

        # Metadata table for information used internally by aerovaldb.
        cur.execute(
            """
            CREATE TABLE metadata(key, value,
            UNIQUE(key))
            """
        )
        self._set_metadata_by_key("created_by", f"aerovaldb_{aerovaldb.__version__}")
        self._set_metadata_by_key(
            "last_modified_by", f"aerovaldb_{aerovaldb.__version__}"
        )

        # Data tables. Currently one table is used per type of asset
        # stored and json blobs are stored in the json column.
        for table_name in AerovalSqliteDB.TABLE_COLUMN_NAMES:
            args = AerovalSqliteDB.TABLE_COLUMN_NAMES[table_name]

            column_names = ",".join(args)
            if table_name in ("reportimages", "mapoverlay"):
                cur.execute(
                    f"""
                    CREATE TABLE IF NOT EXISTS {table_name}(
                        {column_names},
                        ctime TEXT DEFAULT current_timestamp,
                        mtime TEXT DEFAULT current_timestamp,
                        blob BLOB,
                    
                        UNIQUE({column_names})
                    )
                    """
                )
            else:
                cur.execute(
                    f"""
                    CREATE TABLE IF NOT EXISTS {table_name}(
                        {column_names},
                        ctime TEXT DEFAULT current_timestamp,
                        mtime TEXT DEFAULT current_timestamp,
                        json TEXT,

                    UNIQUE({column_names}))
                    """
                )

            cur.execute(
                f"""
                CREATE TRIGGER IF NOT EXISTS update_Timestamp_Trigger_{table_name} AFTER UPDATE On {table_name}
                BEGIN
                   UPDATE {table_name} SET mtime = current_timestamp WHERE rowid = NEW.rowid;
                END;
                """
            )

        self._con.commit()

    def _get_column_list_and_substitution_list(self, kwargs: dict) -> tuple[str, str]:
        keys = list(kwargs.keys())

        columnlist = ", ".join(keys)
        substitutionlist = ", ".join([f":{k}" for k in keys])

        return (columnlist, substitutionlist)

    @override
    async def _get(self, route, route_args, **kwargs):
        cache = kwargs.pop("cache", False)
        default = kwargs.pop("default", None)
        access_type = self._normalize_access_type(kwargs.pop("access_type", None))

        if access_type in [AccessType.FILE_PATH]:
            raise UnsupportedOperation(
                f"sqlitedb does not support access_type FILE_PATH."
            )

        if access_type in [AccessType.URI]:
            return build_uri(route, route_args, kwargs)

        args = route_args | kwargs
        cur = self._con.cursor()
        table_name = await self.TABLE_NAME_LOOKUP.lookup(route, **args)
        args = {
            k: v
            for k, v in args.items()
            if k in AerovalSqliteDB.TABLE_COLUMN_NAMES[table_name]
        }

        columnlist, substitutionlist = self._get_column_list_and_substitution_list(args)
        cur.execute(
            f"""
            SELECT * FROM {table_name}
            WHERE
                ({columnlist}) = ({substitutionlist})
            """,
            args,
        )
        filter_func = self.FILTERS.get(route, None)
        if filter_func:
            filter_vars = {
                k: v
                for k, v in (route_args | kwargs).items()
                if k in inspect.signature(filter_func).parameters.keys()
            }
            if not filter_vars:
                filter_func = None

        try:
            fetched = cur.fetchall()
            if not fetched:
                if default is not None:
                    return default
                # For now, raising a FileNotFoundError, since jsondb does and we want
                # them to be interchangeable. Probably should be a aerovaldb custom
                # exception.
                raise FileNotFoundError("Object not found")
            for r in fetched:
                for k in r.keys():
                    if k in ("json", "blob", "ctime", "mtime"):
                        continue
                    if not (k in route_args | kwargs) and r[k] is not None:
                        break
                else:
                    fetched = r
                    break

        except TypeError as e:
            # Raising file not found error to be consistent with jsondb implementation.
            # Probably should be a custom exception used by aerovaldb.
            raise FileNotFoundError(
                f"No object found for route, {route}, with args {route_args}, {kwargs}"
            ) from e

        json = fetched["json"]
        # No filtered.
        if filter_func is None:
            if access_type == AccessType.JSON_STR:
                return json

            if access_type == AccessType.OBJ:
                dt = simplejson.loads(json, allow_nan=True)

            if access_type == AccessType.MTIME:
                dt = datetime.datetime.strptime(
                    fetched["mtime"], AerovalSqliteDB.SQLITE_TIMESTAMP_FORMAT
                )

            if access_type == AccessType.CTIME:
                dt = datetime.datetime.strptime(
                    fetched["ctime"], AerovalSqliteDB.SQLITE_TIMESTAMP_FORMAT
                )

            return dt

        # Filtered.
        if filter_func is not None:
            if access_type == AccessType.MTIME:
                return datetime.datetime.strptime(
                    fetched["mtime"], AerovalSqliteDB.SQLITE_TIMESTAMP_FORMAT
                )
            if access_type == AccessType.CTIME:
                return datetime.datetime.strptime(
                    fetched["ctime"], AerovalSqliteDB.SQLITE_TIMESTAMP_FORMAT
                )
            obj = simplejson.loads(fetched["json"], allow_nan=True)

            obj = filter_func(obj, **filter_vars)
            if access_type == AccessType.OBJ:
                return obj

            if access_type == AccessType.JSON_STR:
                return json_dumps_wrapper(obj)

        raise UnsupportedOperation

    @override
    async def _put(self, obj, route, route_args, **kwargs):
        cur = self._con.cursor()

        table_name = await self.TABLE_NAME_LOOKUP.lookup(route, **(route_args | kwargs))

        args = route_args | kwargs
        args = {
            k: v
            for k, v in args.items()
            if k in AerovalSqliteDB.TABLE_COLUMN_NAMES[table_name]
        }

        columnlist, substitutionlist = self._get_column_list_and_substitution_list(args)

        json = obj
        if not isinstance(json, str):
            json = json_dumps_wrapper(json)

        args.update(json=json)
        cur.execute(
            f"""
            REPLACE INTO {table_name}({columnlist}, json)
            VALUES({substitutionlist}, :json)
            """,
            args,
        )

        self._set_metadata_by_key(
            "last_modified_by", f"aerovaldb_{aerovaldb.__version__}"
        )
        self._con.commit()

    @async_and_sync
    @override
    async def get_by_uri(
        self,
        uri: str | QueryEntry,
        /,
        access_type: str | AccessType = AccessType.OBJ,
        cache: bool = False,
        default=None,
    ):
        if access_type in [AccessType.URI]:
            return uri

        route, route_args, kwargs = parse_uri(uri)

        if route == Route.REPORT_IMAGE:
            return await self.get_report_image(
                route_args["project"],
                route_args["experiment"],
                route_args["path"],
                access_type=access_type,
            )

        if route == Route.MAP_OVERLAY:
            return await self.get_map_overlay(
                route_args["project"],
                route_args["experiment"],
                route_args["source"],
                route_args["variable"],
                route_args["date"],
                access_type=access_type,
            )

        return await self._get(
            route,
            route_args,
            access_type=access_type,
            cache=cache,
            default=default,
            **kwargs,
        )

    @async_and_sync
    @override
    async def put_by_uri(self, obj, uri: str | QueryEntry):
        route, route_args, kwargs = parse_uri(uri)
        if route == Route.REPORT_IMAGE:
            await self.put_report_image(
                obj, route_args["project"], route_args["experiment"], route_args["path"]
            )
            return

        if route == Route.MAP_OVERLAY:
            await self.put_map_overlay(
                obj,
                route_args["project"],
                route_args["experiment"],
                route_args["source"],
                route_args["variable"],
                route_args["date"],
            )
            return

        await self._put(obj, route, route_args, **kwargs)

    @async_and_sync
    @override
    async def list_all(self):
        cur = self._con.cursor()
        result = []
        for table_name in self.TABLE_COLUMN_NAMES.keys():
            route = AerovalSqliteDB.TABLE_NAME_TO_ROUTE[table_name]
            cur.execute(
                f"""
                SELECT * FROM {table_name}
                """
            )
            fetched = cur.fetchall()

            for r in fetched:
                arg_names = _column_titles_from_route(route)

                route_args = {}
                kwargs = {}
                for k in r.keys():
                    if k in ["json", "blob", "ctime", "mtime"]:
                        continue
                    if k in arg_names:
                        route_args[k] = r[k]
                    else:
                        kwargs[k] = r[k]

                uri = build_uri(route, route_args, kwargs)
                result.append(uri)
        return result

    def _get_lock_file(self) -> str:
        os.makedirs(os.path.expanduser("~/.aerovaldb/.lock/"), exist_ok=True)
        lock_file = os.path.join(
            os.environ.get("AVDB_LOCK_DIR", os.path.expanduser("~/.aerovaldb/.lock/")),
            md5(self._dbfile.encode()).hexdigest(),
        )
        return lock_file

    @override
    def lock(self):
        if self._use_real_lock:
            return FileLock(self._get_lock_file())

        return FakeLock()

    @async_and_sync
    @override
    async def list_timeseries(
        self,
        project: str,
        experiment: str,
        /,
        access_type: str | AccessType = AccessType.URI,
    ):
        if access_type != AccessType.URI:
            raise ValueError(
                f"Invalid access_type. Got {access_type}, expected AccessType.URI"
            )

        cur = self._con.cursor()
        cur.execute(
            f"""
            SELECT * FROM timeseries
            WHERE project=? AND experiment=?
            """,
            (project, experiment),
        )
        fetched = cur.fetchall()

        route = AerovalSqliteDB.TABLE_NAME_TO_ROUTE["timeseries"]
        result = []
        for r in fetched:
            arg_names = _column_titles_from_route(route)
            route_args = {}
            kwargs = {}
            for k in r.keys():
                if k in ["json", "blob", "ctime", "mtime"]:
                    continue

                if k in arg_names:
                    route_args[k] = r[k]
                else:
                    kwargs[k] = r[k]

            uri = build_uri(route, route_args, kwargs)
            result.append(uri)
        return result

    @override
    def rm_experiment_data(self, project: str, experiment: str) -> None:
        cur = self._con.cursor()
        for table in [
            "glob_stats",
            "contour",
            "contour1",
            "timeseries",
            "timeseries_weekly",
            "config",
            "menu",
            "statistics",
            "ranges",
            "regions",
            "models_style0",
            "map0",
            "map1",
            "scatter0",
            "scatter1",
            "profiles",
            "heatmap_timeseries0",
            "heatmap_timeseries1",
            "heatmap_timeseries2",
            "forecast",
            "fairmode",
            "gridded_map",
            "mapoverlay",
        ]:
            cur.execute(
                f"""
                DELETE FROM {table} WHERE project=? AND experiment=?
                """,
                (project, experiment),
            )

    @async_and_sync
    @override
    async def get_report_image(
        self,
        project: str,
        experiment: str,
        path: str,
        access_type: str | AccessType = AccessType.BLOB,
    ):
        access_type = self._normalize_access_type(access_type)

        if access_type not in [AccessType.BLOB, AccessType.MTIME, AccessType.CTIME]:
            raise UnsupportedOperation(
                f"Sqlitedb does not support accesstype {access_type}."
            )

        cur = self._con.cursor()
        cur.execute(
            """
            SELECT * FROM reportimages
            WHERE
                (project, experiment, path) = (?, ?, ?)
            """,
            (project, experiment, path),
        )
        fetched = cur.fetchone()

        if fetched is None:
            raise FileNotFoundError(f"Object not found. {project, experiment, path}")

        if access_type == AccessType.BLOB:
            return fetched["blob"]

        if access_type == AccessType.MTIME:
            return datetime.datetime.strptime(
                fetched["mtime"], AerovalSqliteDB.SQLITE_TIMESTAMP_FORMAT
            )

        if access_type == AccessType.CTIME:
            return datetime.datetime.strptime(
                fetched["ctime"], AerovalSqliteDB.SQLITE_TIMESTAMP_FORMAT
            )

    @async_and_sync
    @override
    async def put_report_image(self, obj, project: str, experiment: str, path: str):
        cur = self._con.cursor()

        if not isinstance(obj, bytes):
            raise TypeError(f"Expected bytes. Got {type(obj)}")

        cur.execute(
            """
            INSERT OR REPLACE INTO reportimages(project, experiment, path, blob)
            VALUES(?, ?, ?, ?)
            """,
            (project, experiment, path, obj),
        )
        self._con.commit()

    @async_and_sync
    @override
    async def get_map_overlay(
        self,
        project: str,
        experiment: str,
        source: str,
        variable: str,
        date: str,
        access_type: str | AccessType = AccessType.BLOB,
    ):
        access_type = self._normalize_access_type(access_type)

        if access_type not in [AccessType.BLOB, AccessType.MTIME, AccessType.CTIME]:
            raise UnsupportedOperation(
                f"Sqlitedb does not support accesstype {access_type}."
            )

        cur = self._con.cursor()
        cur.execute(
            """
            SELECT * FROM mapoverlay
            WHERE
                (project, experiment, source, variable, date) = (?, ?, ?, ?, ?)
            """,
            (project, experiment, source, variable, date),
        )
        fetched = cur.fetchone()

        if fetched is None:
            raise FileNotFoundError(
                f"Object not found. {project, experiment, source, variable, date}"
            )

        if access_type == AccessType.BLOB:
            return fetched["blob"]

        if access_type == AccessType.MTIME:
            return datetime.datetime.strptime(
                fetched["mtime"], AerovalSqliteDB.SQLITE_TIMESTAMP_FORMAT
            )

        if access_type == AccessType.CTIME:
            return datetime.datetime.strptime(
                fetched["ctime"], AerovalSqliteDB.SQLITE_TIMESTAMP_FORMAT
            )

    @async_and_sync
    @override
    async def put_map_overlay(
        self,
        obj,
        project: str,
        experiment: str,
        source: str,
        variable: str,
        date: str,
    ):
        cur = self._con.cursor()

        if not isinstance(obj, bytes):
            raise TypeError(f"Expected bytes. Got {type(obj)}")

        cur.execute(
            """
            INSERT OR REPLACE INTO mapoverlay(project, experiment, source, variable, date, blob)
            VALUES(?, ?, ?, ?, ?, ?)
            """,
            (project, experiment, source, variable, date, obj),
        )
        self._con.commit()

    @async_and_sync
    @override
    async def get_contour(
        self,
        project: str,
        experiment: str,
        obsvar: str,
        model: str,
        /,
        *args,
        timestep: str,
        access_type: str | AccessType = AccessType.OBJ,
        cache: bool = False,
        default=None,
        **kwargs,
    ):
        access_type = self._normalize_access_type(access_type)
        try:
            result = await self._get(
                Route.CONTOUR,
                {
                    "project": project,
                    "experiment": experiment,
                    "obsvar": obsvar,
                    "model": model,
                },
                timestep=timestep,
                access_type=access_type,
                cache=cache,
            )
        except (FileNotFoundError, KeyError):
            pass
        else:
            return result

        try:
            result = await self._get(
                Route.CONTOUR_TIMESPLIT,
                {
                    "project": project,
                    "experiment": experiment,
                    "obsvar": obsvar,
                    "model": model,
                    "timestep": timestep,
                },
                access_type=access_type,
                cache=cache,
            )
        except FileNotFoundError:
            pass
        else:
            return result

        if default is not None:
            return default

        raise FileNotFoundError

    @async_and_sync
    @override
    async def put_contour(
        self,
        obj,
        project: str,
        experiment: str,
        obsvar: str,
        model: str,
        /,
        timestep: str,
        *args,
        **kwargs,
    ):
        if timestep is None:
            logger.warning(
                "Writing contours without providing timestep is deprecated and will be removed in a future release."
            )

            await self._put(
                obj,
                Route.CONTOUR,
                {
                    "project": project,
                    "experiment": experiment,
                    "obsvar": obsvar,
                    "model": model,
                },
            )
            return

        await self._put(
            obj,
            Route.CONTOUR_TIMESPLIT,
            {
                "project": project,
                "experiment": experiment,
                "obsvar": obsvar,
                "model": model,
                "timestep": timestep,
            },
        )

    @async_and_sync
    @override
    async def list_glob_stats(
        self,
        project: str,
        experiment: str,
        /,
        access_type: str | AccessType = AccessType.URI,
    ):
        if access_type != AccessType.URI:
            raise ValueError(
                f"Invalid access_type. Got {access_type}, expected AccessType.URI"
            )

        cur = self._con.cursor()
        cur.execute(
            f"""
            SELECT * FROM glob_stats
            WHERE project=? AND experiment=?
            """,
            (project, experiment),
        )
        fetched = cur.fetchall()

        route = AerovalSqliteDB.TABLE_NAME_TO_ROUTE["glob_stats"]
        result = []
        for r in fetched:
            arg_names = extract_substitutions(route.value)
            route_args = {}
            kwargs = {}
            for k in r.keys():
                if k in ["json", "blob", "ctime", "mtime"]:
                    continue

                if k in arg_names:
                    route_args[k] = r[k]
                else:
                    kwargs[k] = r[k]

            uri = build_uri(route, route_args, kwargs)
            result.append(uri)

        return result



import logging
import sys
from abc import ABC
from typing import Awaitable, Callable, Hashable, Mapping

from packaging.version import Version

if sys.version_info >= (3, 12):
    from typing import override
else:
    from typing_extensions import override

logger = logging.getLogger(__name__)

VersionProvider = Callable[[str, str], Awaitable[Version]]


class SkipMapper(Exception):
    """
    Exception raised when a TemplateMapper does not want to or
    can't handle a request.
    """

    pass


class StringMapper:
    """
    Class for mapping one type of string to the appropriate other
    type of string. It is used by jsonfiledb to map from route to
    the appropriate template string, and in sqlitedb to map from
    route to the correct table name.

    It supports delivering different value strings based on
    additional constraints (such as version).
    """

    def __init__(self, lookup_table: dict, /, version_provider: VersionProvider):
        """
        :param lookup_table : A configuration lookuptable.
        """
        self._lookuptable = lookup_table
        for k in self._lookuptable:
            # A single string will always be returned for that key.
            if isinstance(self._lookuptable[k], str):
                self._lookuptable[k] = ConstantMapper(self._lookuptable[k])

            # Make sure value is a list.
            if not isinstance(self._lookuptable[k], list):
                self._lookuptable[k] = [self._lookuptable[k]]

            # If stringlist of len > 1, treat it as a priority order.
            if len(self._lookuptable[k]) > 1 and all(
                [isinstance(x, str) for x in self._lookuptable[k]]
            ):
                self._lookuptable[k] = [PriorityMapper(self._lookuptable[k])]

        self._version_provider = version_provider

    def __iter__(self):
        return iter(self._lookuptable.keys())

    async def lookup(self, key: Hashable, **kwargs) -> str:
        """
        Performs a lookup of the value for the given key.

        :param key : The key for which to lookup a value.
        :param kwargs : Additional values which will be used for constraint
        matching.
        :raises KeyError
            If no entry exists for the key in the lookup table.
        :return
            The looked up string value.
        """
        if (version := kwargs.pop("version", None)) is None:
            version_provider = self._version_provider
        else:

            async def version_helper(p, e):
                return Version(version)

            version_provider = version_helper

        try:
            values = self._lookuptable[key]
        except KeyError as e:
            raise KeyError(f"Key '{key}' does not exist in lookup table.") from e

        kwargs = kwargs | {"version_provider": version_provider}
        return_value = None
        for v in values:
            try:
                return_value = await v(**kwargs)
            except SkipMapper:
                continue

            break

        if not return_value:
            raise ValueError(f"No valid value found for key '{key}'")

        return return_value


class Mapper(ABC):
    """
    This class is a base class for objects that implement a
    file path template selection algorithm. Inheriting
    implementations should implement the __call_ function,
    and raising SkipMapper if the implementation can't or
    won't handle the request.
    """

    async def __call__(self, *args, **kwargs) -> str:
        raise NotImplementedError


class VersionConstraintMapper(Mapper):
    """
    This class returns its provided template if the data version read
    from a config file matches the configured bounds of this class.

    Version is matched according to the following equation:
        min_version <= version < max_version
    """

    def __init__(
        self,
        template: str,
        *,
        min_version: str | None = None,
        max_version: str | None = None,
    ):
        """
        :param template : The template string to return.
        :min_version : The minimum version to which to apply this template (inclusive).
        :max_version : The maximum version to which to apply this template (exclusive).
        :version_provider :
            Function or other callable that takes 'project' and 'experiment', returning
            the Pyaerocom version that wrote the data to be read.
        """
        self.min_version = None
        self.max_version = None

        if min_version is not None:
            self.min_version = Version(min_version)
        if max_version is not None:
            self.max_version = Version(max_version)

        self.template = template

    @override
    async def __call__(self, *args, **kwargs) -> str:
        version_provider = kwargs.pop("version_provider")
        version = await version_provider(kwargs["project"], kwargs["experiment"])
        if not version:
            raise ValueError("No version provided")
        logger.debug(f"Trying template string {self.template}")

        if self.min_version is not None and version < self.min_version:
            logging.debug(
                f"Skipping due to version mismatch. {version} < {self.min_version}"
            )
            raise SkipMapper
        if self.max_version is not None and version >= self.max_version:
            logging.debug(
                f"Skipping due to version mismatch. {version} >= {self.max_version}"
            )
            raise SkipMapper

        return self.template


class PriorityMapper(Mapper):
    """
    This class takes a list of templates, trying them in turn
    and returning the first template that fits the provided
    parameters.
    """

    def __init__(self, templates: list[str] | dict):
        """
        :param templates
            If list, a list of template strings which will
            be matched in order.
            If dict, a mapping between the template string
            and an alternative string to match against is
            expected (i. e. {"test", "{a}"} would try matching
            against "{a}" during lookup, but return "test" if
            match.)
        """
        if isinstance(templates, list):
            self.templates = templates
            self.match = templates
        elif isinstance(templates, dict):
            self.templates = []
            self.match = []
            for k, v in templates.items():
                self.templates.append(k)
                self.match.append(v)

    @override
    async def __call__(self, *args, **kwargs) -> str:
        selected_template = None
        for t, m in zip(self.templates, self.match):
            try:
                m.format(**kwargs)
                selected_template = t
                break
            except:
                continue

        if selected_template is None:
            raise SkipMapper

        return selected_template


class ConstantMapper(Mapper):
    def __init__(self, template: str):
        self.template = template

    @override
    async def __call__(self, *args, **kwargs) -> str:
        return self.template



from packaging.version import Version

from ..routes import Route


# The motivation for doing this is explained here:
# https://github.com/metno/aerovaldb/issues/119
# tl;dr: it works around ambiguity in parsing of legacy
# filenames, to be able to extract metadata from file
# names reliably.
def post_process_args(
    route: Route, args: dict[str, str], kwargs: dict[str, str], *, version: Version
) -> tuple[dict[str, str], dict[str, str]]:
    if route == Route.MAP:
        args, kwargs = _post_process_maps_args_kwargs(args, kwargs)
    elif route in [Route.TIMESERIES, Route.TIMESERIES_WEEKLY]:
        args, kwargs = _post_process_timeseries_args_kwargs(
            args, kwargs, version=version
        )
    elif route == Route.HEATMAP_TIMESERIES:
        args, kwargs = _post_process_heatmap_ts_args_kwargs(
            args, kwargs, version=version
        )
    elif route == Route.FORECAST:
        args, kwargs = _post_process_forecast_args_kwargs(args, kwargs)
    elif route == Route.SCATTER:
        args, kwargs = _post_process_scatter_args_kwargs(args, kwargs)
    return args, kwargs


def _post_process_maps_args_kwargs(
    args: dict[str, str], kwargs: dict[str, str]
) -> tuple[dict[str, str], dict[str, str]]:
    if "-" in args["obsvar"]:
        splt = args["obsvar"].split("-")

        args["obsvar"] = splt[-1]
        args["network"] = args["network"] + f"-{'-'.join(splt[:-1])}"

    if "-" in args["modvar"]:
        splt = args["modvar"].split("-")

        args["modvar"] = splt[-1]
        args["model"] = args["model"] + f"-{'-'.join(splt[:-1])}"

    return args, kwargs


def _post_process_timeseries_args_kwargs(
    args: dict[str, str], kwargs: dict[str, str], *, version
) -> tuple[dict[str, str], dict[str, str]]:
    if version >= Version("0.29.0.dev1"):
        return args, kwargs

    if "-" in args["obsvar"]:
        splt = args["obsvar"].split("-")
        args["obsvar"] = splt[-1]

        args["network"] = args["network"] + f"-{'-'.join(splt[:-1])}"

    if "_" in args["network"]:
        splt = args["network"].split("_")
        args["network"] = splt[-1]

        args["location"] = args["location"] + f"_{'_'.join(splt[:-1])}"

    return args, kwargs


def _post_process_scatter_args_kwargs(
    args: dict[str, str], kwargs: dict[str, str]
) -> tuple[dict[str, str], dict[str, str]]:
    if "-" in args["obsvar"]:
        splt = args["obsvar"].split("-")

        args["obsvar"] = splt[-1]
        args["network"] = args["network"] + f"-{'-'.join(splt[:-1])}"

    return args, kwargs


def _post_process_heatmap_ts_args_kwargs(
    args: dict[str, str], kwargs: dict[str, str], *, version
) -> tuple[dict[str, str], dict[str, str]]:
    if version >= Version("0.29.0.dev1"):
        return args, kwargs
    if version <= Version("0.12.2"):
        return args, kwargs
    if version <= Version("0.13.2"):
        if "-" in kwargs["obsvar"]:
            splt: list[str] = kwargs["obsvar"].split("-")
            kwargs["obsvar"] = splt[-1]
            kwargs["network"] = kwargs["network"] + f"-{'-'.join(splt[:-1])}"

        return args, kwargs

    string = "-".join(
        [kwargs["region"], kwargs["network"], kwargs["obsvar"], kwargs["layer"]]
    )
    splt = string.split("-")

    # We know that neither layer nor obsvar can contain -.
    kwargs["layer"] = splt[-1]
    del splt[-1]
    kwargs["obsvar"] = splt[-1]
    del splt[-1]

    # Region can contain - but not _ (otherwise becomes impossible to disambiguate).
    kwargs["region"] = splt[0]
    del splt[0]

    kwargs["network"] = "-".join(splt)

    return args, kwargs


def _post_process_forecast_args_kwargs(
    args: dict[str, str], kwargs: dict[str, str]
) -> tuple[dict[str, str], dict[str, str]]:
    if "-" in args["obsvar"]:
        splt = args["obsvar"].split("-")

        args["obsvar"] = splt[-1]

        args["network"] = args["network"] + f"-{'-'.join(splt[:-1])}"

    return args, kwargs



import datetime
import glob
import importlib.metadata
import inspect
import logging
import os
import shutil
import sys
from hashlib import md5
from pathlib import Path
from typing import Any, Awaitable, Callable, Iterable

import filetype  # type: ignore
import simplejson  # type: ignore
from packaging.version import Version

from aerovaldb.aerovaldb import AerovalDB
from aerovaldb.const import IMG_FILE_EXTS
from aerovaldb.types import AccessType

from ..exceptions import UnsupportedOperation
from ..lock import FakeLock, FileLock
from ..routes import *
from ..utils import (
    async_and_sync,
    build_uri,
    extract_substitutions,
    json_dumps_wrapper,
    parse_formatted_string,
    parse_uri,
    str_to_bool,
)
from ..utils.filter import filter_heatmap, filter_map, filter_regional_stats
from ..utils.string_mapper import StringMapper, VersionConstraintMapper
from .cache import CacheMissError, KeyCacheDecorator, LRUFileCache

if sys.version_info >= (3, 12):
    from typing import override
else:
    from typing_extensions import override

from async_lru import alru_cache

from ..utils.encode import DecodedStr, EncodedStr, decode_str, encode_str
from ..utils.query import QueryEntry
from .backwards_compatibility import post_process_args

logger = logging.getLogger(__name__)


class _LiteralArg(str):
    """Custom string instance that behaves identical to a regular string.
    It is only used in internal isinstance checks to decide whether to
    apply character encoding to a provided arg when constructing a file
    name. This is in order to allow for args which represent a path to
    contain a '/' character which is normally encoded.
    """

    pass


class AerovalJsonFileDB(AerovalDB):
    # Character mapping used for encoding and decoding string values in file names.
    # Note: Order matters for correct encoding, and % should always be the last entry in this dict.
    FNAME_ENCODE_CHARS = {"/": "%1", "_": "%2", "%": "%0"}

    def __init__(self, basedir: str | Path):
        """
        :param basedir The root directory where aerovaldb will look for files.
        """
        self._use_real_lock = str_to_bool(
            os.environ.get("AVDB_USE_LOCKING", ""), default=False
        )
        logger.info(
            f"Initializing aerovaldb for '{basedir}' with locking {self._use_real_lock}"
        )

        self._cache = KeyCacheDecorator(LRUFileCache(max_size=64), max_size=512)

        self._basedir = os.path.abspath(basedir)

        if not os.path.exists(self._basedir):
            os.makedirs(self._basedir)

        self.PATH_LOOKUP = StringMapper(
            {
                Route.HEATMAP: "./{project}/{experiment}/hm/glob_stats_{frequency}.json",
                Route.GLOB_STATS: "./{project}/{experiment}/hm/glob_stats_{frequency}.json",
                Route.REGIONAL_STATS: "./{project}/{experiment}/hm/glob_stats_{frequency}.json",
                # For MAP_OVERLAY, extension is excluded but it will be appended after the fact.
                Route.MAP_OVERLAY: "./{project}/{experiment}/overlay/{variable}_{source}/{variable}_{source}_{date}",
                Route.CONTOUR: "./{project}/{experiment}/contour/{obsvar}_{model}.geojson",
                Route.CONTOUR_TIMESPLIT: "./{project}/{experiment}/contour/{obsvar}_{model}/{obsvar}_{model}_{timestep}.geojson",
                Route.TIMESERIES_WEEKLY: [
                    VersionConstraintMapper(
                        "./{project}/{experiment}/ts/diurnal/{location}_{network}_{obsvar}_{layer}.json",
                        min_version="0.29.0.dev1",
                    ),
                    VersionConstraintMapper(
                        "./{project}/{experiment}/ts/diurnal/{location}_{network}-{obsvar}_{layer}.json",
                        max_version="0.29.0.dev1",
                    ),
                ],
                Route.TIMESERIES: [
                    VersionConstraintMapper(
                        "./{project}/{experiment}/ts/{location}_{network}_{obsvar}_{layer}.json",
                        min_version="0.29.0.dev1",
                    ),
                    VersionConstraintMapper(
                        "./{project}/{experiment}/ts/{location}_{network}-{obsvar}_{layer}.json",
                        max_version="0.29.0.dev1",
                    ),
                ],
                Route.EXPERIMENTS: "./{project}/experiments.json",
                Route.CONFIG: "./{project}/{experiment}/cfg_{project}_{experiment}.json",
                Route.MENU: "./{project}/{experiment}/menu.json",
                Route.STATISTICS: "./{project}/{experiment}/statistics.json",
                Route.RANGES: "./{project}/{experiment}/ranges.json",
                Route.REGIONS: "./{project}/{experiment}/regions.json",
                Route.MODELS_STYLE: [
                    "./{project}/{experiment}/models-style.json",
                    "./{project}/models-style.json",
                ],
                Route.MAP: [
                    VersionConstraintMapper(
                        "./{project}/{experiment}/map/{network}_{obsvar}_{layer}_{model}_{modvar}_{time}.json",
                        min_version="0.29.0.dev1",
                    ),
                    VersionConstraintMapper(
                        "./{project}/{experiment}/map/{network}-{obsvar}_{layer}_{model}-{modvar}_{time}.json",
                        min_version="0.13.2",
                        max_version="0.29.0.dev1",
                    ),
                    VersionConstraintMapper(
                        "./{project}/{experiment}/map/{network}-{obsvar}_{layer}_{model}-{modvar}.json",
                        max_version="0.13.2",
                    ),
                ],
                Route.SCATTER: [
                    VersionConstraintMapper(
                        "./{project}/{experiment}/scat/{network}_{obsvar}_{layer}_{model}_{modvar}_{time}.json",
                        min_version="0.29.0.dev1",
                    ),
                    VersionConstraintMapper(
                        "./{project}/{experiment}/scat/{network}-{obsvar}_{layer}_{model}-{modvar}_{time}.json",
                        min_version="0.13.2",
                        max_version="0.29.0.dev1",
                    ),
                    VersionConstraintMapper(
                        "./{project}/{experiment}/scat/{network}-{obsvar}_{layer}_{model}-{modvar}.json",
                        max_version="0.13.2",
                    ),
                ],
                Route.PROFILES: "./{project}/{experiment}/profiles/{location}_{network}_{obsvar}.json",
                Route.HEATMAP_TIMESERIES: [
                    VersionConstraintMapper(
                        "./{project}/{experiment}/hm/ts/{region}_{network}_{obsvar}_{layer}.json",
                        min_version="0.29.0.dev1",
                    ),
                    VersionConstraintMapper(
                        "./{project}/{experiment}/hm/ts/{region}-{network}-{obsvar}-{layer}.json",
                        min_version="0.13.2",  # https://github.com/metno/pyaerocom/blob/4478b4eafb96f0ca9fd722be378c9711ae10c1f6/setup.cfg
                        max_version="0.29.0.dev1",
                    ),
                    VersionConstraintMapper(
                        "./{project}/{experiment}/hm/ts/{network}-{obsvar}-{layer}.json",
                        min_version="0.12.2",
                        max_version="0.13.2",
                    ),
                    VersionConstraintMapper(
                        "./{project}/{experiment}/hm/ts/stats_ts.json",
                        max_version="0.12.2",
                    ),
                ],
                Route.FORECAST: [
                    VersionConstraintMapper(
                        "./{project}/{experiment}/forecast/{region}_{network}_{obsvar}_{layer}.json",
                        min_version="0.29.0.dev1",
                    ),
                    VersionConstraintMapper(
                        "./{project}/{experiment}/forecast/{region}_{network}-{obsvar}_{layer}.json",
                        max_version="0.29.0.dev1",
                    ),
                ],
                Route.FAIRMODE: "./{project}/{experiment}/fairmode/{region}_{network}_{obsvar}_{layer}.json",
                Route.GRIDDED_MAP: "./{project}/{experiment}/contour/{obsvar}_{model}.json",
                Route.REPORT: "./reports/{project}/{experiment}/{title}.json",
                Route.REPORT_IMAGE: "./reports/{project}/{experiment}/{path}",
            },
            version_provider=self._get_version,
        )

        self.FILTERS: dict[Route, Callable[..., Awaitable[Any]]] = {
            Route.REGIONAL_STATS: filter_regional_stats,
            Route.HEATMAP: filter_heatmap,
            Route.MAP: filter_map,
        }

    def _load_json(
        self,
        key,
        *,
        access_type: AccessType = AccessType.OBJ,
        cache: bool = False,
    ):
        if access_type in [
            AccessType.BLOB,
            AccessType.URI,
            AccessType.FILE_PATH,
            AccessType.MTIME,
            AccessType.CTIME,
        ]:
            ValueError(f"Unable to load json with access_type={access_type}.")

        json_str = self._cache.get(key, bypass_cache=not cache)
        if access_type == AccessType.JSON_STR:
            return json_str

        elif access_type == AccessType.OBJ:
            return simplejson.loads(json_str, allow_nan=True)

        raise UnsupportedOperation(f"{access_type}")

    @async_and_sync
    @alru_cache(maxsize=2048)
    async def _get_version(
        self, project: DecodedStr, experiment: DecodedStr
    ) -> Version:
        """
        Returns the version of pyaerocom used to generate the files for a given project
        and experiment.

        :param project : Project ID.
        :param experiment : Experiment ID.

        :return : A Version object.
        """
        try:
            config = await self.get_config(project, experiment)
        except FileNotFoundError:
            try:
                # If pyaerocom is installed in the current environment, but no config has
                # been written, we use the version of the installed pyaerocom. This is
                # important for tests to work correctly, and for files to be written
                # correctly if the config file happens to be written after data files.
                version = Version(importlib.metadata.version("pyaerocom"))
            except importlib.metadata.PackageNotFoundError:
                version = Version("0.0.1")
            finally:
                return version

        try:
            version_str = config["exp_info"]["pyaerocom_version"]
            version = Version(version_str)
        except KeyError:
            version = Version("0.0.1")

        return version

    @async_and_sync
    async def _get_template(
        self, route: Route, substitutions: dict[str, DecodedStr]
    ) -> str:
        """
        Loops through each instance of TemplateMapper finding the
        appropriate template to use give an route, and a dictionary
        of substitutions.

        :param route : The route for which to look up the template.
        :param substitutions : Dictionary of format substitutions available.

        :returns The template string.

        :raises TemplateNotFound :
            If no valid template was found.
        """
        return await self.PATH_LOOKUP.lookup(route, **substitutions)

    def _prepare_substitutions(self, subs: dict[str, _LiteralArg | str]) -> dict:
        """Prepares template substitutions for inclusion in file path. This mainly
        entails file name encoding.

        :param subs: Dict of subs to be prepared.

        :return: Dict with same keys with the prepared values.
        """
        return {
            k: v
            if isinstance(v, _LiteralArg)
            else encode_str(v, encode_chars=AerovalJsonFileDB.FNAME_ENCODE_CHARS)
            for k, v in subs.items()
        }

    @override
    async def _get(
        self,
        route,
        route_args,
        **kwargs,
    ):
        use_caching = kwargs.pop("cache", False)
        default = kwargs.pop("default", None)
        _raise_file_not_found_error = kwargs.pop("_raise_file_not_found_error", True)
        access_type = self._normalize_access_type(kwargs.pop("access_type", None))

        path_template = await self._get_template(route, (route_args | kwargs))
        logger.debug(f"Using template string {path_template}")

        substitutions = self._prepare_substitutions(route_args | kwargs)

        logger.debug(f"Fetching data for {route}.")

        relative_path = path_template.format(**substitutions)

        file_path = os.path.join(self._basedir, relative_path)
        logger.debug(f"Fetching file {file_path} as {access_type}-")

        filter_func = self.FILTERS.get(route, None)
        if filter_func:
            filter_vars = {
                k: v
                for k, v in (route_args | kwargs).items()
                if k in inspect.signature(filter_func).parameters.keys()
            }
            if not filter_vars:
                filter_func = None

        if not os.path.exists(file_path):
            if default is None or access_type == AccessType.FILE_PATH:
                if _raise_file_not_found_error:
                    raise FileNotFoundError(f"File {file_path} does not exist.")
                else:
                    return file_path
            return default

        if filter_func is None:
            if access_type == AccessType.FILE_PATH:
                return file_path
            if access_type == AccessType.URI:
                return build_uri(route, route_args, kwargs)

            if access_type == AccessType.MTIME:
                return datetime.datetime.fromtimestamp(os.path.getmtime(file_path))
            if access_type == AccessType.CTIME:
                return datetime.datetime.fromtimestamp(os.path.getctime(file_path))

            return self._load_json(
                file_path, access_type=access_type, cache=use_caching
            )

        if access_type == AccessType.FILE_PATH:
            raise UnsupportedOperation("Filtered endpoints can not return a filepath")

        if access_type == AccessType.MTIME:
            return datetime.datetime.fromtimestamp(os.path.getmtime(file_path))
        if access_type == AccessType.CTIME:
            return datetime.datetime.fromtimestamp(os.path.getctime(file_path))

        filter_params = [kwargs[k] for k in sorted(kwargs.keys())]
        key = f"{file_path}::{'/'.join(filter_params)}"

        try:
            obj = self._load_json(key, access_type=AccessType.OBJ, cache=use_caching)
        except CacheMissError:
            obj = self._load_json(file_path)
            obj = filter_func(obj, **filter_vars)
            if use_caching:
                self._cache.put(json_dumps_wrapper(obj), key=key)

        if access_type == AccessType.OBJ:
            return obj

        if access_type == AccessType.JSON_STR:
            return json_dumps_wrapper(obj)

        raise UnsupportedOperation

    @override
    async def _put(self, obj, route, route_args, **kwargs):
        """Jsondb implemention of database put operation.

        If obj is string, it is assumed to be a wellformatted json string.
        Otherwise it is assumed to be a serializable python object.
        """
        path_template = await self._get_template(route, route_args | kwargs)

        assert all(
            isinstance(v, DecodedStr | str) for v in (route_args | kwargs).values()
        )
        substitutions = self._prepare_substitutions(route_args | kwargs)

        relative_path = path_template.format(**substitutions)

        file_path = str(Path(os.path.join(self._basedir, relative_path)).resolve())

        logger.debug(f"Mapped route {route} / { route_args} to file {file_path}.")

        if not os.path.exists(os.path.dirname(file_path)):
            os.makedirs(os.path.dirname(file_path))
        if isinstance(obj, str):
            json = obj
        else:
            json = json_dumps_wrapper(obj)
        with open(file_path, "w") as f:
            f.write(json)

    @override
    def rm_experiment_data(self, project: str, experiment: str) -> None:
        """Deletes ALL data associated with an experiment.

        :param project : Project ID.
        :param experiment : Experiment ID.
        """
        exp_dir = os.path.join(self._basedir, project, experiment)

        if os.path.exists(exp_dir):
            logger.info(
                f"Removing experiment data for project {project}, experiment, {experiment}."
            )
            shutil.rmtree(exp_dir)

    @async_and_sync
    @override
    async def get_regional_stats(
        self,
        project: str,
        experiment: str,
        frequency: str,
        network: str,
        variable: str,
        layer: str,
        /,
        *args,
        **kwargs,
    ):
        """Fetches regional stats from the database.

        :param project: The project ID.
        :param experiment: The experiment ID.
        :param frequency: The frequency.
        :param network: Observation network.
        :param variable: Variable name.
        """
        return await self._get(
            Route.REGIONAL_STATS,
            {"project": project, "experiment": experiment, "frequency": frequency},
            access_type=kwargs.get("access_type", AccessType.OBJ),
            network=network,
            variable=variable,
            layer=layer,
            cache=True,
        )

    @async_and_sync
    @override
    async def get_heatmap(
        self,
        project: str,
        experiment: str,
        frequency: str,
        region: str,
        time: str,
        /,
        *args,
        **kwargs,
    ):
        """Fetches heatmap data from the database

        :param project: The project ID.
        :param experiment: The experiment ID.
        :param frequency: The frequency.
        :param region: Region.
        :param time: Time.
        """
        return await self._get(
            Route.HEATMAP,
            {"project": project, "experiment": experiment, "frequency": frequency},
            access_type=kwargs.get("access_type", AccessType.OBJ),
            region=region,
            time=time,
            cache=True,
        )

    @async_and_sync
    @alru_cache(maxsize=1000)
    async def _get_query_entry_for_file(self, file_path: str) -> QueryEntry:
        """
        For the provided data file path, returns the corresponding
        URI.

        :param file_path : The file_path.
        """
        file_path = os.path.join(self._basedir, file_path)
        file_path = os.path.relpath(file_path, start=self._basedir)

        _, ext = os.path.splitext(file_path)

        if "/overlay/" in file_path:
            file_path = str(Path(file_path).parent / Path(file_path).stem)

        if file_path.startswith("reports/") and ext.lower() in IMG_FILE_EXTS:
            # TODO: Fix this.
            # The image endpoint is the only endpoint which needs to accept an arbitrary path
            # under the experiment directory. Treating it as a special case for now.
            split = file_path.split("/")
            project = split[1]
            experiment = split[2]
            path = "/".join(split[3:])
            uri = build_uri(
                Route.REPORT_IMAGE,
                {
                    "project": decode_str(
                        project, encode_chars=AerovalJsonFileDB.FNAME_ENCODE_CHARS
                    ),
                    "experiment": decode_str(
                        experiment, encode_chars=AerovalJsonFileDB.FNAME_ENCODE_CHARS
                    ),
                    "path": path,
                },
                {},
            )
            entry = QueryEntry(
                uri,
                Route.REPORT_IMAGE,
                {"project": project, "experiment": experiment, "path": path},
            )
            return entry

        for route in self.PATH_LOOKUP._lookuptable:
            if not (route == Route.MODELS_STYLE):
                if file_path.startswith("reports/"):
                    _str = "/".join(file_path.split("/")[1:3])
                    subs = parse_formatted_string("{project}/{experiment}", _str)
                else:
                    _str = "/".join(file_path.split("/")[0:2])
                    subs = parse_formatted_string("{project}/{experiment}", _str)
            else:
                try:
                    subs = parse_formatted_string(
                        "{project}/{experiment}/models-style.json", file_path
                    )
                except Exception:
                    try:
                        subs = parse_formatted_string(
                            "{project}/models-style.json", file_path
                        )
                    except:
                        continue

            subs = {
                k: decode_str(v, encode_chars=AerovalJsonFileDB.FNAME_ENCODE_CHARS)
                for k, v in subs.items()
            }
            template = await self._get_template(route, subs)

            if "experiment" in subs:
                version = await self._get_version(subs["project"], subs["experiment"])
            else:
                # Project level models style does not have a version because version is defined
                # per experiment. version doesn't matter for models-style because it is priority
                # based, so we set a dummy value to simplify.
                version = Version("0.0.1")
            route_arg_names = extract_substitutions(route.value)

            try:
                all_args = parse_formatted_string(template, f"./{file_path}")  # type: ignore
                route_args = {k: v for k, v in all_args.items() if k in route_arg_names}
                kwargs = {
                    k: v for k, v in all_args.items() if not (k in route_arg_names)
                }
                route_args, kwargs = post_process_args(
                    route, route_args, kwargs, version=version
                )
                route_args = {
                    k: decode_str(v, encode_chars=AerovalJsonFileDB.FNAME_ENCODE_CHARS)
                    for k, v in route_args.items()
                }
                kwargs = {
                    k: decode_str(v, encode_chars=AerovalJsonFileDB.FNAME_ENCODE_CHARS)
                    for k, v in kwargs.items()
                }
            except Exception:
                continue
            else:
                uri = build_uri(route, route_args, kwargs | {"version": str(version)})
                entry = QueryEntry(uri, Route(route), route_args | kwargs)
                return entry

        raise ValueError(f"Unable to build URI for file path {file_path}")

    @async_and_sync
    @override
    async def list_timeseries(
        self,
        project: str,
        experiment: str,
    ):
        logger.warning("list_all is deprecated. Please consider using query() instead.")
        return [
            uri.uri
            for uri in await self.query(
                Route.TIMESERIES, project=project, experiment=experiment
            )
        ]

    @async_and_sync
    @override
    async def list_map(
        self,
        project: str,
        experiment: str,
    ):
        logger.warning("list_all is deprecated. Please consider using query() instead.")
        return [
            uri.uri
            for uri in await self.query(
                Route.MAP, project=project, experiment=experiment
            )
        ]

    @async_and_sync
    @override
    async def get_by_uri(
        self,
        uri: str | QueryEntry,
        /,
        access_type: str | AccessType = AccessType.OBJ,
        cache: bool = False,
        default=None,
    ):
        access_type = self._normalize_access_type(access_type)
        if access_type in [AccessType.URI]:
            return uri

        route, route_args, kwargs = parse_uri(uri)

        if route.value.startswith("/v0/report-image/"):
            return await self.get_report_image(
                route_args["project"],
                route_args["experiment"],
                route_args["path"],
                access_type=access_type,
            )

        if route.value.startswith("/v0/map-overlay/"):
            return await self.get_map_overlay(
                route_args["project"],
                route_args["experiment"],
                route_args["source"],
                route_args["variable"],
                route_args["date"],
                access_type=access_type,
            )

        return await self._get(
            route,
            route_args,
            cache=cache,
            default=default,
            access_type=access_type,
            **kwargs,
        )

    @async_and_sync
    @override
    async def put_by_uri(self, obj, uri: str | QueryEntry):
        route, route_args, kwargs = parse_uri(uri)

        if route.value.startswith("/v0/report-image/"):
            await self.put_report_image(
                obj, route_args["project"], route_args["experiment"], route_args["path"]
            )
            return

        if route.value.startswith("/v0/map-overlay/"):
            await self.put_map_overlay(
                obj,
                route_args["project"],
                route_args["experiment"],
                route_args["source"],
                route_args["variable"],
                route_args["date"],
            )
            return

        await self._put(obj, route, route_args, **kwargs)

    def _get_lock_file(self) -> str:
        os.makedirs(os.path.expanduser("~/.aerovaldb/lock/"), exist_ok=True)
        lock_file = os.path.join(
            os.environ.get("AVDB_LOCK_DIR", os.path.expanduser("~/.aerovaldb/lock/")),
            md5(self._basedir.encode()).hexdigest(),
        )
        return lock_file

    @override
    def lock(self):
        if self._use_real_lock:
            return FileLock(self._get_lock_file())

        return FakeLock()

    @async_and_sync
    @override
    async def query(
        self, asset_type: Route | Iterable[Route] | None = None, **kwargs
    ) -> list[QueryEntry]:
        if asset_type is None:
            asset_type = set(Route)
        elif isinstance(asset_type, Route):
            asset_type = set([asset_type])
        elif isinstance(asset_type, Iterable):
            asset_type = set(asset_type)
        else:
            raise TypeError(f"Expected Route | Iterable[Route]. Got {type(asset_type)}")

        glb = glob.iglob(
            os.path.join(glob.escape(self._basedir), "./**"), recursive=True
        )

        result = []
        for f in glb:
            if os.path.isfile(f):
                try:
                    entry = await self._get_query_entry_for_file(f)
                except (ValueError, KeyError):
                    continue
                else:
                    if entry.type in asset_type:
                        if all(entry.meta[k] == v for k, v in kwargs.items()):
                            result.append(entry)

        return result

    @async_and_sync
    @override
    async def list_all(self):
        logger.warning("list_all is deprecated. Please consider using query() instead.")
        return [uri.uri for uri in await self.query()]

    @async_and_sync
    @override
    async def get_report_image(
        self,
        project: str,
        experiment: str,
        path: str,
        access_type: str | AccessType = AccessType.BLOB,
    ):
        access_type = self._normalize_access_type(access_type)

        if access_type not in (
            AccessType.FILE_PATH,
            AccessType.BLOB,
            AccessType.MTIME,
            AccessType.CTIME,
        ):
            raise UnsupportedOperation(
                f"The report image endpoint does not support access type {access_type}."
            )

        if access_type in (AccessType.MTIME, AccessType.CTIME):
            return await self._get(
                route=Route.REPORT_IMAGE,
                route_args={
                    "project": project,
                    "experiment": experiment,
                    "path": _LiteralArg(path),
                },
                access_type=access_type,
            )
        file_path = await self._get(
            route=Route.REPORT_IMAGE,
            route_args={
                "project": project,
                "experiment": experiment,
                "path": _LiteralArg(path),
            },
            access_type=AccessType.FILE_PATH,
        )
        logger.debug(f"Fetching image with path '{file_path}'")

        if access_type == AccessType.FILE_PATH:
            return file_path

        with open(file_path, "rb") as f:
            return f.read()

    @async_and_sync
    @override
    async def put_report_image(self, obj, project: str, experiment: str, path: str):
        template = await self._get_template(Route.REPORT_IMAGE, {})

        file_path = os.path.join(
            self._basedir,
            template.format(project=project, experiment=experiment, path=path),
        )
        os.makedirs(Path(file_path).parent, exist_ok=True)
        with open(file_path, "wb") as f:
            f.write(obj)

    @async_and_sync
    @override
    async def get_map_overlay(
        self,
        project: str,
        experiment: str,
        source: str,
        variable: str,
        date: str,
        access_type: str | AccessType = AccessType.BLOB,
    ):
        access_type = self._normalize_access_type(access_type)

        if access_type not in (
            AccessType.FILE_PATH,
            AccessType.BLOB,
            AccessType.MTIME,
            AccessType.CTIME,
        ):
            raise UnsupportedOperation(
                f"The report image endpoint does not support access type {access_type}."
            )

        for ext in IMG_FILE_EXTS:
            file_path = await self._get(
                route=Route.MAP_OVERLAY,
                route_args={
                    "project": project,
                    "experiment": experiment,
                    "source": source,
                    "variable": variable,
                    "date": date,
                },
                _raise_file_not_found_error=False,
                access_type=AccessType.FILE_PATH,
            )

            file_path += ext
            if os.path.exists(file_path):
                break
        else:
            raise FileNotFoundError(
                f"Overlay for {project}/{experiment}/{source}/{variable}/{date} does not exist."
            )

        logger.debug(f"Fetching image with path '{file_path}'")

        if access_type in [AccessType.MTIME]:
            return datetime.datetime.fromtimestamp(os.path.getmtime(file_path))
        if access_type in [AccessType.CTIME]:
            return datetime.datetime.fromtimestamp(os.path.getctime(file_path))

        if access_type == AccessType.FILE_PATH:
            return file_path

        with open(file_path, "rb") as f:
            return f.read()

    @async_and_sync
    @override
    async def put_map_overlay(
        self,
        obj,
        project: str,
        experiment: str,
        source: str,
        variable: str,
        date: str,
    ):
        """Putter for map overlay images.

        :param obj : The object to be stored.
        :param project : Project ID.
        :param experiment : Experiment ID.
        :param source : Data source. Can be either an observation network or a model ID.
        :param variable : Variable name.
        :param date : Date.
        """
        template = await self._get_template(Route.MAP_OVERLAY, {})

        subs = self._prepare_substitutions(
            {
                "project": project,
                "experiment": experiment,
                "source": source,
                "variable": variable,
                "date": date,
            }
        )
        file_path = os.path.join(
            self._basedir,
            template.format(**subs),
        )

        ext = filetype.guess_extension(obj)
        if ext is None:
            raise ValueError(
                f"Could not guess image file extension of provided image data starting with '0x{obj[:20].hex()}'."
            )
        file_path += f".{ext}"

        Path(file_path).parent.mkdir(exist_ok=True, parents=True)
        with open(file_path, "wb") as f:
            f.write(obj)

    @async_and_sync
    @override
    async def get_contour(
        self,
        project: str,
        experiment: str,
        obsvar: str,
        model: str,
        /,
        *args,
        timestep: str,
        access_type: str | AccessType = AccessType.OBJ,
        cache: bool = False,
        default=None,
        **kwargs,
    ):
        access_type = self._normalize_access_type(access_type)

        try:
            file_path = await self._get(
                Route.CONTOUR,
                {
                    "project": project,
                    "experiment": experiment,
                    "obsvar": obsvar,
                    "model": model,
                },
                # timestep=timestep,
                access_type=AccessType.FILE_PATH,
                cache=True,
            )
            try:
                if timestep is None:
                    key = file_path
                else:
                    key = f"{file_path}::{timestep}"

                result = simplejson.loads(self._cache.get(key), allow_nan=True)
            except CacheMissError:
                result = await self._get(
                    Route.CONTOUR,
                    {
                        "project": project,
                        "experiment": experiment,
                        "obsvar": obsvar,
                        "model": model,
                    },
                    access_type=AccessType.OBJ,
                    cache=True,
                )
                if timestep is not None:
                    for t, value in result.items():
                        self._cache.put(
                            json_dumps_wrapper(value), key=f"{file_path}::{t}"
                        )
                    result = result[timestep]
        except (FileNotFoundError, KeyError):
            pass
        else:
            if access_type == AccessType.OBJ:
                return result
            if access_type == AccessType.JSON_STR:
                return json_dumps_wrapper(result)

        try:
            result = await self._get(
                Route.CONTOUR_TIMESPLIT,
                {
                    "project": project,
                    "experiment": experiment,
                    "obsvar": obsvar,
                    "model": model,
                    "timestep": timestep,
                },
                access_type=access_type,
                cache=cache,
            )
        except FileNotFoundError:
            pass
        else:
            return result

        if default is not None:
            return default

        raise FileNotFoundError

    @async_and_sync
    @override
    async def put_contour(
        self,
        obj,
        project: str,
        experiment: str,
        obsvar: str,
        model: str,
        /,
        timestep: str | None = None,
        *args,
        **kwargs,
    ):
        if timestep is None:
            logger.warning(
                "Writing contours without providing timestep is deprecated and will be removed in a future release."
            )

            await self._put(
                obj,
                Route.CONTOUR,
                {
                    "project": project,
                    "experiment": experiment,
                    "obsvar": obsvar,
                    "model": model,
                },
            )
            return

        await self._put(
            obj,
            Route.CONTOUR_TIMESPLIT,
            {
                "project": project,
                "experiment": experiment,
                "obsvar": obsvar,
                "model": model,
                "timestep": timestep,
            },
        )

    @async_and_sync
    @override
    async def rm_by_uri(self, uri: str | QueryEntry):
        file_path = await self.get_by_uri(str(uri), access_type=AccessType.FILE_PATH)
        logger.debug("Removing file '%s'.", file_path)

        if os.path.exists(file_path):
            os.remove(file_path)

    @async_and_sync
    @override
    async def list_glob_stats(
        self,
        project: str,
        experiment: str,
        /,
        access_type: str | AccessType = AccessType.URI,
    ) -> list[str]:
        logger.warning("list_glob_stats() is deprecated. Please use query instead.")

        # Route.HEATMAP below is intentional, as this maintains old behaviour for compatibility.
        # Will be removed in future since the name is misleading (it returns heatmap while being
        # named list_glob_stats).
        return [
            uri.uri
            for uri in await self.query(
                Route.HEATMAP, project=project, experiment=experiment
            )
        ]

    @async_and_sync
    @override
    async def get_report(
        self,
        project: str,
        experiment: str,
        title: str,
        /,
        *args,
        access_type: str | AccessType = AccessType.OBJ,
        cache: bool = False,
        default=None,
        **kwargs,
    ):
        return await self._get(
            Route.REPORT,
            {
                "project": _LiteralArg(project),
                "experiment": _LiteralArg(experiment),
                "title": _LiteralArg(title),
            },
            access_type=access_type,
            cache=cache,
            default=default,
        )

    @async_and_sync
    @override
    async def put_report(
        self, obj, project: str, experiment: str, title: str, /, *args, **kwargs
    ):
        """Store report.

        :param obj: The object to be stored.
        :param project: Project ID.
        :param experiment: Experiment ID.
        :param title: Report title (ie. filename without extension).
        """
        await self._put(
            obj,
            Route.REPORT,
            {
                "project": _LiteralArg(project),
                "experiment": _LiteralArg(experiment),
                "title": _LiteralArg(title),
            },
            **kwargs,
        )



from importlib import metadata

__version__ = metadata.version(__package__)

from .aerovaldb import AerovalDB
from .exceptions import *
from .plugins import list_engines, open
from .routes import Route
from .types import *



import re
import urllib

from aerovaldb.utils.query import QueryEntry

from ..routes import Route
from .encode import decode_str, encode_str

# Note: Order matters for correct encoding, and % should always be the last entry in this dict.
encode_chars = {"/": "%1", "%": "%0"}


def encode_arg(string: str):
    """Encodes a string to work in an aerovaldb URI.

    :param string: The string to be encoded.
    :return: Encoded string.
    """
    return encode_str(string, encode_chars=encode_chars)


def decode_arg(string: str):
    """Decodes an encoded string to work in an aerovaldb URI. Inverse of
    encode_arg

    :param string: The string to be decoded.
    :return: Decoded string.
    """
    return decode_str(string, encode_chars=encode_chars)


def extract_substitutions(template: str):
    """
    For a python template string, extracts the names between curly brackets:

    For example 'blah blah {test} blah {test2}' returns ["test", "test2"]
    """
    return re.findall(r"\{([a-zA-Z-]*?)\}", template)


def parse_formatted_string(
    template: str, string: str, *, force_split: list[str] | None = ["/"]
):
    """Parse formatted string. Meant to be the inverse of str.format()

    :param template: Template string.
    :param string: String to be matched.
    :param force_split: Optional list of single character strings which will force a break between tokens.
    :raises Exception: If unable to match `s` against template.
    :return: Dict of extracted arguments.

    Limitations
    -----------
    - Only works for format strings that use the named curly bracket notation.
    In other words no %s or {} notation.

    Example:
    >>> from aerovaldb.utils.uri import parse_formatted_string
    >>> parse_formatted_string("{a}/{b}", "test1/test2")
    {'a': 'test1', 'b': 'test2'}
    """
    if force_split is None:
        force_split = []

    if any([not isinstance(x, str) for x in force_split]):
        raise TypeError(
            f"force_split got elements that aren't string. Got {force_split}."
        )

    if any([len(x) != 1 for x in force_split]):
        raise ValueError(
            f"force_split must be a list of single character strings. Got {force_split}."
        )

    original_string = string
    keywords = extract_substitutions(template)

    pattern = "(" + "|".join([re.escape("{" + k + "}") for k in keywords]) + ")"
    segments = [x for x in re.split(pattern, template) if x != ""]
    # Segments is a list of constant strings and keywords (Keywords starting with '{').
    # For instance 'a{b}c{d}' -> ['a', '{b}', 'c', '{d}']

    result = {}
    while len(segments) > 0:
        token = segments[0]
        next_token = None
        if len(segments) >= 2:
            next_token = segments[1]
        if token.startswith("{"):
            # Token is a keyword, so try to extract it.
            ls: list[str] = []
            if next_token is not None:
                if next_token.startswith("{"):
                    raise Exception(
                        f"Two successive keywords can not be disambiguated (s='{original_string}; template='{template}')"
                    )

                # First opportunity where the remainder of the string starts with the next token is where we stop matching.
                # Note: This prevents some strings from being matched if the next token is also part of the string that should
                # be matched, but it isn't causing problems for now.
                while len(ls) < len(string) and not (
                    string[len(ls) :].startswith(next_token)
                ):
                    char = string[len(ls)]
                    if char in force_split:
                        break
                    ls.append(char)

                extr = "".join(ls)
            else:
                extr = string

            result[token.replace("{", "").replace("}", "")] = extr
            string = string[len(extr) :]
        else:
            if not string.startswith(token):
                break

            string = string[len(token) :]

        segments = segments[1:]
    if len(segments) > 0:
        raise Exception(
            f"Formatted string '{original_string}' did not match template string '{template}'"
        )
    return result


def parse_uri(uri: str | QueryEntry) -> tuple[Route, dict[str, str], dict[str, str]]:
    """
    Parses an uri returning a tuple consisting of
    - The route against which it was matched.
    - Route args.
    - Additional kwargs.

    Parameters
    ----------
    uri :
        The uri to be parsed.

    Example
    -------
    >>> from aerovaldb.utils.uri import parse_uri
    >>> parse_uri('/v0/experiments/project')
    (<Route.EXPERIMENTS: '/v0/experiments/{project}'>, {'project': 'project'}, {})
    """
    uri = str(uri)

    split = uri.split("?")

    for template in [x.value for x in Route]:
        if len(split) == 1:
            try:
                route_args = parse_formatted_string(template, split[0])
            except Exception:
                continue
            else:
                for k, v in route_args.items():
                    route_args[k] = decode_arg(v)
                return (Route(template), route_args, dict())

        elif len(split) == 2:
            try:
                route_args = parse_formatted_string(template, split[0])
            except Exception:
                continue

            kwargs = urllib.parse.parse_qs(split[1])  # type: ignore
            kwargs = {k: v[0] for k, v in kwargs.items()}

            for k, v in route_args.items():
                route_args[k] = decode_arg(v)
            for k, v in kwargs.items():
                kwargs[k] = decode_arg(v)
            return (Route(template), route_args, kwargs)

    raise ValueError(f"URI {uri} is not a valid URI.")


def build_uri(route: Route, route_args: dict, kwargs: dict = {}) -> str:
    for k, v in route_args.items():
        route_args[k] = encode_arg(v)
    for k, v in kwargs.items():
        kwargs[k] = encode_arg(v)
    uri = route.value.format(**route_args)
    if kwargs:
        queries = "&".join([f"{k}={v}" for k, v in kwargs.items()])
        uri = f"{uri}?{queries}"

    return uri
